from __future__ import annotations

import pytest

from xdsl.ir import ParametrizedAttribute, Data
from xdsl.irdl import irdl_attr_definition, builder
from xdsl.parser import BaseParser
from xdsl.printer import Printer
from xdsl.utils.exceptions import BuilderNotFoundException


@irdl_attr_definition
class NoBuilderAttr(ParametrizedAttribute):
    name = "test.no_builder_attr"


def test_no_builder_default():
    attr = NoBuilderAttr.build(NoBuilderAttr())
    assert attr == NoBuilderAttr()


def test_no_builder_exception():
    with pytest.raises(BuilderNotFoundException):
        NoBuilderAttr.build(3)


@irdl_attr_definition
class OneBuilderAttr(Data[str]):
    name = "test.one_builder_attr"

    @staticmethod
    @builder
    def from_int(data: int) -> OneBuilderAttr:
        return OneBuilderAttr(str(data))

    @staticmethod
    def parse_parameter(parser: BaseParser) -> str:
        raise NotImplementedError()

    @staticmethod
    def print_parameter(data: str, printer: Printer) -> None:
        raise NotImplementedError()


def test_one_builder_default():
    attr = OneBuilderAttr.build(OneBuilderAttr("a"))
    assert attr == OneBuilderAttr("a")


def test_one_builder_builder():
    attr = OneBuilderAttr.build(1)
    assert attr == OneBuilderAttr("1")


def test_one_builder_exception():
    with pytest.raises(BuilderNotFoundException):
        OneBuilderAttr.build("1")


def test_one_builder_extra_params():
    with pytest.raises(BuilderNotFoundException):
        OneBuilderAttr.build(1, 2)


@irdl_attr_definition
class OneBuilderTwoArgsAttr(Data[str]):
    name = "test.one_builder_attr"

    @staticmethod
    @builder
    def from_int(data1: int, data2: str) -> OneBuilderAttr:
        return OneBuilderAttr(str(data1) + data2)

    @staticmethod
    def parse_parameter(parser: BaseParser) -> str:
        raise NotImplementedError()

    @staticmethod
    def print_parameter(data: str, printer: Printer) -> None:
        raise NotImplementedError()


def test_one_builder_two_args_not_enough_args():
    with pytest.raises(BuilderNotFoundException):
        OneBuilderTwoArgsAttr.build(1)


def test_one_builder_two_args_extra_args():
    with pytest.raises(BuilderNotFoundException):
        OneBuilderTwoArgsAttr.build(1, "a", "b")


@irdl_attr_definition
class TwoBuildersAttr(Data[str]):
    name = "test.two_builder_attr"

    @staticmethod
    @builder
    def from_int(data: int) -> TwoBuildersAttr:
        return TwoBuildersAttr(str(data))

    @staticmethod
    @builder
    def from_str(s: str) -> TwoBuildersAttr:
        return TwoBuildersAttr(s)

    @staticmethod
    def parse_parameter(parser: BaseParser) -> str:
        raise NotImplementedError()

    @staticmethod
    def print_parameter(data: str, printer: Printer) -> None:
        raise NotImplementedError()


def test_two_builders_default():
    attr = TwoBuildersAttr.build(TwoBuildersAttr("a"))
    assert attr == TwoBuildersAttr("a")


def test_two_builders_first_builder():
    attr = TwoBuildersAttr.build(1)
    assert attr == TwoBuildersAttr("1")


def test_two_builders_second_builder():
    attr = TwoBuildersAttr.build("1")
    assert attr == TwoBuildersAttr("1")


def test_two_builders_bad_args():
    with pytest.raises(BuilderNotFoundException):
        TwoBuildersAttr.build([])


@irdl_attr_definition
class BuilderDefaultArgAttr(Data[str]):
    name = "test.builder_default_arg_attr"

    @staticmethod
    @builder
    def from_int(data1: int,
                 data2: int = 42,
                 data3: int = 17) -> BuilderDefaultArgAttr:
        return BuilderDefaultArgAttr(f"{data1}, {data2}, {data3}")

    @staticmethod
    def parse_parameter(parser: BaseParser) -> str:
        raise NotImplementedError()

    @staticmethod
    def print_parameter(data: str, printer: Printer) -> None:
        raise NotImplementedError()


def test_builder_default_args_missing_args():
    with pytest.raises(BuilderNotFoundException):
        BuilderDefaultArgAttr.build()


def test_builder_default_args_only_required_args():
    attr = BuilderDefaultArgAttr.build(4)
    assert attr == BuilderDefaultArgAttr("4, 42, 17")


def test_builder_default_args_only_required_args_and_one():
    attr = BuilderDefaultArgAttr.build(4, 5)
    assert attr == BuilderDefaultArgAttr("4, 5, 17")


def test_builder_default_args_all_args():
    attr = BuilderDefaultArgAttr.build(4, 5, 6)
    assert attr == BuilderDefaultArgAttr("4, 5, 6")


def test_builder_default_args_extra_args():
    with pytest.raises(BuilderNotFoundException):
        BuilderDefaultArgAttr.build(4, 5, 6, 7)


@irdl_attr_definition
class BuilderUnionArgAttr(Data[str]):
    name = "test.builder_union_arg_attr"

    @staticmethod
    @builder
    def from_int(data: str | int) -> BuilderUnionArgAttr:
        return BuilderUnionArgAttr(str(data))

    @staticmethod
    def parse_parameter(parser: BaseParser) -> str:
        raise NotImplementedError()

    @staticmethod
    def print_parameter(data: str, printer: Printer) -> None:
        raise NotImplementedError()


def test_builder_union_arg_first():
    attr = BuilderUnionArgAttr.build(4)
    assert attr == BuilderUnionArgAttr("4")


def test_builder_union_arg_second():
    attr = BuilderUnionArgAttr.build("4")
    assert attr == BuilderUnionArgAttr("4")


def test_builder_union_arg_bad_argument():
    with pytest.raises(BuilderNotFoundException):
        BuilderUnionArgAttr.build([])



from __future__ import annotations

from typing import Annotated, List

from xdsl.ir import SSAValue, Block, Region, Operation, Attribute, Dialect
from xdsl.irdl import (VarOpResult, VarOperand, irdl_op_definition, Operand,
                       AnyAttr)
from xdsl.dialects.builtin import IntegerType


@irdl_op_definition
class If(Operation):
    name: str = "scf.if"
    output: Annotated[VarOpResult, AnyAttr()]
    cond: Annotated[Operand, IntegerType.from_width(1)]

    true_region: Region
    # TODO this should be optional under certain conditions
    false_region: Region

    @staticmethod
    def get(cond: SSAValue | Operation, return_types: List[Attribute],
            true_region: Region | List[Block] | List[Operation],
            false_region: Region | List[Block] | List[Operation]) -> If:
        return If.build(operands=[cond],
                        result_types=[return_types],
                        regions=[true_region, false_region])


@irdl_op_definition
class Yield(Operation):
    name: str = "scf.yield"
    arguments: Annotated[VarOperand, AnyAttr()]

    @staticmethod
    def get(*operands: SSAValue | Operation) -> Yield:
        return Yield.create(
            operands=[SSAValue.get(operand) for operand in operands])


@irdl_op_definition
class Condition(Operation):
    name: str = "scf.condition"
    cond: Annotated[Operand, IntegerType.from_width(1)]
    arguments: Annotated[VarOperand, AnyAttr()]

    @staticmethod
    def get(cond: SSAValue | Operation,
            *output_ops: SSAValue | Operation) -> Condition:
        return Condition.build(
            operands=[cond, [output for output in output_ops]])


@irdl_op_definition
class While(Operation):
    name: str = "scf.while"
    arguments: Annotated[VarOperand, AnyAttr()]

    res: Annotated[VarOpResult, AnyAttr()]
    before_region: Region
    after_region: Region

    # TODO verify dependencies between scf.condition, scf.yield and the regions
    def verify_(self):
        for (idx, arg) in enumerate(self.arguments):
            if self.before_region.blocks[0].args[idx].typ != arg.typ:
                raise Exception(
                    f"Block arguments with wrong type, expected {arg.typ}, "
                    f"got {self.before_region.blocks[0].args[idx].typ}")

        for (idx, res) in enumerate(self.res):
            if self.after_region.blocks[0].args[idx].typ != res.typ:
                raise Exception(
                    f"Block arguments with wrong type, expected {res.typ}, "
                    f"got {self.after_region.blocks[0].args[idx].typ}")

    @staticmethod
    def get(operands: List[SSAValue | Operation],
            result_types: List[Attribute],
            before: Region | List[Operation] | List[Block],
            after: Region | List[Operation] | List[Block]) -> While:
        op = While.build(operands=operands,
                         result_types=result_types,
                         regions=[before, after])
        return op


Scf = Dialect([If, Yield, Condition, While], [])



from __future__ import annotations

import xdsl.dialects.builtin as builtin

from typing import Any, Callable, Generic, TypeAlias, TypeVar, Literal
from xdsl.dialects.builtin import Signedness


class _FrontendType:
    """Represents any type in the frontend."""

    @staticmethod
    def to_xdsl() -> Callable[..., Any]:
        raise NotImplementedError()


# Type parameters for integers.
_Width = TypeVar("_Width", bound=int, covariant=True)
_Signedness = TypeVar("_Signedness", bound=Signedness, covariant=True)


class _Integer(Generic[_Width, _Signedness], _FrontendType):
    """Represents an integer type in the frontend. Should not be used explicitly."""

    @staticmethod
    def to_xdsl() -> Callable[..., Any]:
        return builtin.IntegerType.from_width


# Type aliases for signless integers.
i1: TypeAlias = _Integer[Literal[1], Literal[Signedness.SIGNLESS]]
i32: TypeAlias = _Integer[Literal[32], Literal[Signedness.SIGNLESS]]
i64: TypeAlias = _Integer[Literal[64], Literal[Signedness.SIGNLESS]]

# Type aliases for signed integers.
si32: TypeAlias = _Integer[Literal[32], Literal[Signedness.SIGNED]]
si64: TypeAlias = _Integer[Literal[64], Literal[Signedness.SIGNED]]

# Type aliases for unsigned integers.
ui32: TypeAlias = _Integer[Literal[32], Literal[Signedness.UNSIGNED]]
ui64: TypeAlias = _Integer[Literal[64], Literal[Signedness.UNSIGNED]]


class _Index(_FrontendType):
    """Represents an index type in the frontend. Should not be used explicitly."""

    @staticmethod
    def to_xdsl() -> Callable[..., Any]:
        return builtin.IndexType


# Type alias for index type.
index: TypeAlias = _Index


class _Float16(_FrontendType):
    """Represents a 16-bit floating-point type in the frontend. Should not be used explicitly."""

    @staticmethod
    def to_xdsl() -> Callable[..., Any]:
        return builtin.Float16Type


class _Float32(_FrontendType):
    """Represents a 32-bit floating-point type in the frontend. Should not be used explicitly."""

    @staticmethod
    def to_xdsl() -> Callable[..., Any]:
        return builtin.Float32Type


class _Float64(_FrontendType):
    """Represents a 64-bit floating-point type in the frontend. Should not be used explicitly."""

    @staticmethod
    def to_xdsl() -> Callable[..., Any]:
        return builtin.Float64Type


# Type alias for floating-point types.
f16: TypeAlias = _Float16
f32: TypeAlias = _Float32
f64: TypeAlias = _Float64



import pytest

from typing import Annotated

from xdsl.dialects.builtin import i32, StringAttr
from xdsl.dialects.arith import Constant

from xdsl.ir import Block, Operation, OpResult
from xdsl.irdl import irdl_op_definition, OpResult


def test_ssa():
    a = OpResult(i32, [], [])
    c = Constant.from_int_and_width(1, i32)

    with pytest.raises(TypeError):
        _ = a.get([c])

    b0 = Block.from_ops([c])
    with pytest.raises(TypeError):
        _ = a.get(b0)


@irdl_op_definition
class TwoResultOp(Operation):
    name: str = "test.tworesults"

    res1: Annotated[OpResult, StringAttr]
    res2: Annotated[OpResult, StringAttr]


def test_var_mixed_builder():
    op = TwoResultOp.build(result_types=[0, 2])
    b = OpResult(i32, [], [])

    with pytest.raises(ValueError):
        _ = b.get(op)



import ast
import pytest

from xdsl.frontend.exception import CodeGenerationException
from xdsl.frontend.python_code_check import CheckAndInlineConstants


def test_const_correctly_evaluated_I():

    src = \
"""
a: Const[i32] = 2 ** 5
x = a
"""
    stmts = ast.parse(src).body
    CheckAndInlineConstants.run(stmts)
    assert ast.unparse(stmts).endswith("x = 32")


def test_const_correctly_evaluated_II():

    src = \
"""
a: Const[i32] = 4
x: i64 = a + 2
"""
    stmts = ast.parse(src).body
    CheckAndInlineConstants.run(stmts)
    assert ast.unparse(stmts).endswith("x: i64 = 4 + 2")


def test_const_correctly_evaluated_III():

    src = \
"""
a: Const[i32] = 4
b: Const[i32] = len([1, 2, 3, 4])
x: Const[i32] = a + b
y = x
"""
    stmts = ast.parse(src).body
    CheckAndInlineConstants.run(stmts)
    assert ast.unparse(stmts).endswith("y = 8")


def test_const_correctly_evaluated_IV():

    src = \
"""
a: Const[i32] = 4
def foo(y: i32):
    x: i32 = a + y
"""
    stmts = ast.parse(src).body
    CheckAndInlineConstants.run(stmts)
    assert ast.unparse(stmts).endswith("x: i32 = 4 + y")


def test_const_correctly_evaluated_V():

    src = \
"""
a: Const[i32] = 4
b: Const[i32] = 4
def foo(y: i32):
    c: Const[i32] = a + b + 2
    x: i32 = c
"""
    stmts = ast.parse(src).body
    CheckAndInlineConstants.run(stmts)
    assert ast.unparse(stmts).endswith("x: i32 = 10")


def test_raises_exception_on_assignemnt_to_const_I():
    src = \
"""
a: Const[i32] = 2 ** 5
a = 34
"""
    stmts = ast.parse(src).body
    with pytest.raises(CodeGenerationException) as err:
        CheckAndInlineConstants.run(stmts)
    assert err.value.msg == "Constant 'a' is already defined and cannot be assigned to."


def test_raises_exception_on_assignemnt_to_const_II():
    src = \
"""
x: Const[i32] = 100
def foo():
    x: i32 = 2
    return
"""
    stmts = ast.parse(src).body
    with pytest.raises(CodeGenerationException) as err:
        CheckAndInlineConstants.run(stmts)
    assert err.value.msg == "Constant 'x' is already defined."


def test_raises_exception_on_assignemnt_to_const_III():
    src = \
"""
y: Const[i32] = 100
@block
def bb0():
    y = 32
    return
bb0()
"""
    stmts = ast.parse(src).body
    with pytest.raises(CodeGenerationException) as err:
        CheckAndInlineConstants.run(stmts)
    assert err.value.msg == "Constant 'y' is already defined and cannot be assigned to."


def test_raises_exception_on_assignemnt_to_const_IV():
    src = \
"""
z: Const[i32] = 100
def foo(x: i32):
    @block
    def bb0(z: i32):
        return
    bb0(x)
"""
    stmts = ast.parse(src).body
    with pytest.raises(CodeGenerationException) as err:
        CheckAndInlineConstants.run(stmts)
    assert err.value.msg == "Constant 'z' is already defined and cannot be used as a function/block argument name."


def test_raises_exception_on_duplicate_const():
    src = \
"""
z: Const[i32] = 100
z: Const[i32] = 2
"""
    stmts = ast.parse(src).body
    with pytest.raises(CodeGenerationException) as err:
        CheckAndInlineConstants.run(stmts)
    assert err.value.msg == "Constant 'z' is already defined."


def test_raises_exception_on_evaluation_error_I():
    src = \
"""
z: Const[i32] = 23 / 0
"""
    stmts = ast.parse(src).body
    with pytest.raises(CodeGenerationException) as err:
        CheckAndInlineConstants.run(stmts)
    assert err.value.msg == "Non-constant expression cannot be assigned to constant variable 'z' or cannot be evaluated."


def test_raises_exception_on_evaluation_error_II():
    src = \
"""
a: Const[i32] = x + 12
"""
    stmts = ast.parse(src).body
    with pytest.raises(CodeGenerationException) as err:
        CheckAndInlineConstants.run(stmts)
    assert err.value.msg == "Non-constant expression cannot be assigned to constant variable 'a' or cannot be evaluated."



from __future__ import annotations
from typing import Annotated, List, Union

from xdsl.dialects.builtin import StringAttr, FunctionType, FlatSymbolRefAttr
from xdsl.ir import SSAValue, Operation, Block, Region, Attribute, Dialect
from xdsl.irdl import (VarOpResult, irdl_op_definition, VarOperand, AnyAttr,
                       OpAttr, OptOpAttr)
from xdsl.utils.exceptions import VerifyException


@irdl_op_definition
class FuncOp(Operation):
    name: str = "func.func"

    body: Region
    sym_name: OpAttr[StringAttr]
    function_type: OpAttr[FunctionType]
    sym_visibility: OptOpAttr[StringAttr]

    def verify_(self) -> None:
        # TODO: how to verify that there is a terminator?
        entry_block: Block = self.body.blocks[0]
        block_arg_types = [arg.typ for arg in entry_block.args]
        if self.function_type.inputs.data != block_arg_types:
            raise VerifyException(
                "Expected entry block arguments to have the same types as the function input types"
            )

    @staticmethod
    def from_callable(name: str, input_types: List[Attribute],
                      return_types: List[Attribute],
                      func: Block.BlockCallback) -> FuncOp:
        type_attr = FunctionType.from_lists(input_types, return_types)
        attributes = {
            "sym_name": name,
            "function_type": type_attr,
            "sym_visibility": "private"
        }
        op = FuncOp.build(attributes=attributes,
                          regions=[
                              Region.from_block_list(
                                  [Block.from_callable(input_types, func)])
                          ])
        return op

    @staticmethod
    def from_region(name: str, input_types: List[Attribute],
                    return_types: List[Attribute], region: Region) -> FuncOp:
        type_attr = FunctionType.from_lists(input_types, return_types)
        attributes = {
            "sym_name": name,
            "function_type": type_attr,
            "sym_visibility": "private"
        }
        op = FuncOp.build(attributes=attributes, regions=[region])
        return op


@irdl_op_definition
class Call(Operation):
    name: str = "func.call"
    arguments: Annotated[VarOperand, AnyAttr()]
    callee: OpAttr[FlatSymbolRefAttr]

    # Note: naming this results triggers an ArgumentError
    res: Annotated[VarOpResult, AnyAttr()]
    # TODO how do we verify that the types are correct?

    @staticmethod
    def get(callee: Union[str, FlatSymbolRefAttr],
            operands: List[Union[SSAValue, Operation]],
            return_types: List[Attribute]) -> Call:
        return Call.build(operands=[operands],
                          result_types=[return_types],
                          attributes={"callee": callee})


@irdl_op_definition
class Return(Operation):
    name: str = "func.return"
    arguments: Annotated[VarOperand, AnyAttr()]

    def verify_(self) -> None:
        func_op = self.parent_op()
        assert isinstance(func_op, FuncOp)

        function_return_types = func_op.function_type.outputs.data
        return_types = [arg.typ for arg in self.arguments]
        if function_return_types != return_types:
            raise VerifyException(
                "Expected arguments to have the same types as the function output types"
            )

    @staticmethod
    def get(*ops: Union[Operation, SSAValue]) -> Return:
        ops = [op for op in ops] if ops != () else []
        return Return.build(operands=[ops])


Func = Dialect([FuncOp, Call, Return], [])



import ast


def block(*params):
    """
    Decorator used to mark function as a basic block.
    ```
    def foo(a: int) -> int:
        @block
        def bb0(x: int):
            y: int = x + 2
            bb1(y)
        
        @block
        def bb1(z: int):
            return z
        
        # Entry-point.
        bb0(a)
    ```
    """

    def decorate(*params):
        return None

    return decorate


def is_block(node: ast.FunctionDef) -> bool:
    return len(node.decorator_list) == 1 and isinstance(
        name := node.decorator_list[0], ast.Name) and name.id == "block"



from typing import Annotated

from pytest import raises

from xdsl.ir import MLContext
from xdsl.irdl import AnyAttr, irdl_op_definition, Operation, VarOperand, VarOpResult
from xdsl.parser import XDSLParser
from xdsl.utils.exceptions import ParseError


@irdl_op_definition
class UnkownOp(Operation):
    name = "unknown"
    ops: Annotated[VarOperand, AnyAttr()]
    res: Annotated[VarOpResult, AnyAttr()]


def check_error(prog: str, line: int, column: int, message: str):
    ctx = MLContext()
    ctx.register_op(UnkownOp)

    parser = XDSLParser(ctx, prog)
    with raises(ParseError) as e:
        parser.parse_operation()

    assert e.value.span

    for err in e.value.history.iterate():
        if message in err.error.msg:
            assert err.error.span.get_line_col() == (line, column)
            break
    else:
        assert False, "'{}' not found in an error message {}!".format(
            message, e.value.args)


def test_parser_missing_equal():
    """Test a missing equal sign error."""
    ctx = MLContext()
    ctx.register_op(UnkownOp)

    prog = \
"""
unknown() {
  %0 : !i32 unknown()
}
"""
    check_error(prog, 3, 12,
                "Operation definitions expect an `=` after op-result-list!")


def test_parser_redefined_value():
    """Test an SSA value redefinition error."""
    ctx = MLContext()
    ctx.register_op(UnkownOp)

    prog = \
"""
unknown() {
  %val : !i32 = unknown()
  %val : !i32 = unknown()
}
"""
    check_error(prog, 4, 2, "SSA value %val is already defined")


def test_parser_missing_operation_name():
    """Test a missing operation name error."""
    ctx = MLContext()
    ctx.register_op(UnkownOp)

    prog = \
"""
unknown() {
  %val : !i32 = 
}
"""
    check_error(prog, 4, 0, "Expected an operation name here")


def test_parser_malformed_type():
    """Test a missing type error."""
    ctx = MLContext()
    ctx.register_op(UnkownOp)

    prog = \
"""
unknown() {
  %val : i32 = unknown()
}
"""
    check_error(prog, 3, 9, "Expected type of value-id here!")



import re
from io import StringIO
from typing import Annotated
from xdsl.ir import (Attribute, Data, MLContext, MLIRType, Operation,
                     ParametrizedAttribute, Region)
from xdsl.irdl import (AnyAttr, ParameterDef, VarOpResult, VarOperand,
                       irdl_attr_definition, irdl_op_definition)
from xdsl.parser import BaseParser, XDSLParser
from xdsl.printer import Printer


@irdl_op_definition
class ModuleOp(Operation):
    """Module operation. Redefined to not depend on the builtin dialect."""
    name = "module"
    region: Region


@irdl_op_definition
class AnyOp(Operation):
    """Operation only used for testing."""
    name = "any"
    op: Annotated[VarOperand, AnyAttr()]
    res: Annotated[VarOpResult, AnyAttr()]


@irdl_attr_definition
class DataAttr(Data[int]):
    """Attribute only used for testing."""
    name = "data_attr"

    @staticmethod
    def parse_parameter(parser: BaseParser) -> int:
        return parser.parse_int_literal()

    @staticmethod
    def print_parameter(data: int, printer: Printer) -> None:
        printer.print(data)


@irdl_attr_definition
class DataType(Data[int], MLIRType):
    """Attribute only used for testing."""
    name = "data_type"

    @staticmethod
    def parse_parameter(parser: BaseParser) -> int:
        return parser.parse_int_literal()

    @staticmethod
    def print_parameter(data: int, printer: Printer) -> None:
        printer.print(data)


@irdl_attr_definition
class ParamAttr(ParametrizedAttribute):
    name = "param_attr"


@irdl_attr_definition
class ParamAttrWithParam(ParametrizedAttribute):
    name = "param_attr_with_param"
    data: ParameterDef[Attribute]


@irdl_attr_definition
class ParamType(ParametrizedAttribute, MLIRType):
    name = "param_type"


@irdl_attr_definition
class ParamAttrWithCustomFormat(ParametrizedAttribute):
    name = "param_custom_format"
    param1: ParameterDef[ParamAttr]

    def print_parameters(self, printer: Printer) -> None:
        printer.print("~~")


def print_as_mlir_and_compare(test_prog: str, expected: str):
    ctx = MLContext()

    ctx.register_op(ModuleOp)
    ctx.register_op(AnyOp)
    ctx.register_attr(DataAttr)
    ctx.register_attr(DataType)
    ctx.register_attr(ParamAttr)
    ctx.register_attr(ParamType)
    ctx.register_attr(ParamAttrWithParam)
    ctx.register_attr(ParamAttrWithCustomFormat)

    parser = XDSLParser(ctx, test_prog)
    module = parser.parse_operation()

    res = StringIO()
    printer = Printer(target=Printer.Target.MLIR, stream=res)
    printer.print_op(module)

    # Remove all whitespace from the expected string.
    regex = re.compile(r'[^\S]+')
    assert (regex.sub("", res.getvalue()).strip() == \
            regex.sub("", expected).strip())


def test_empty_op():
    """Test printing an empty operation."""
    print_as_mlir_and_compare(
        """any()""",
        """"any"() : () -> ()""",
    )


def test_data_attr():
    """Test printing an operation with a data attribute."""
    print_as_mlir_and_compare(
        """any() [ "attr" = !data_attr<42> ]""",
        """"any"() {"attr" = #data_attr<42>} : () -> ()""",
    )


def test_data_type():
    """Test printing an operation with a data type."""
    print_as_mlir_and_compare(
        """%0 : !data_type<42> = any()""",
        """%0 = "any"() : () -> !data_type<42>""",
    )


def test_param_attr():
    """Test printing an operation with a parametrized attribute."""
    print_as_mlir_and_compare(
        """any() [ "attr" = !param_attr ]""",
        """"any"() {"attr" = #param_attr } : () -> ()""",
    )


def test_param_type():
    """Test printing an operation with a parametrized type."""
    print_as_mlir_and_compare(
        """%0 : !param_type = any()""",
        """%0 = "any"() : () -> !param_type""",
    )


def test_param_attr_with_param():
    """
    Test printing an operation with a parametrized attribute with parameters.
    """
    print_as_mlir_and_compare(
        """any() [ "attr" = !param_attr_with_param<!param_attr> ]""",
        """"any"() {"attr" = #param_attr_with_param<#param_attr> }
          : () -> ()""",
    )

    print_as_mlir_and_compare(
        """any() [ "attr" = !param_attr_with_param<!param_type> ]""",
        """"any"() {"attr" = #param_attr_with_param<!param_type> }
          : () -> ()""",
    )


def test_op_with_region():
    """Test printing an operation with a region."""

    print_as_mlir_and_compare(
        """module() {}""",
        """"module"() ({}) : () -> ()""",
    )


def test_op_with_results():
    """Test printing an operation with results."""

    print_as_mlir_and_compare(
        """%0 : !param_attr = any()""",
        """%0 = "any"() : () -> #param_attr""",
    )

    print_as_mlir_and_compare(
        """%0 : !param_attr, %1 : !param_type = any()""",
        """%0, %1 = "any"() : () -> (#param_attr, !param_type)""",
    )


def test_op_with_operands():
    """Test printing an operation with operands."""
    print_as_mlir_and_compare(
        """module() {
           %0 : !param_attr = any()
           any(%0 : !param_attr)
        }""",
        """"module"() ({
              %0 = "any"() : () -> #param_attr
              "any"(%0) : (#param_attr) -> ()
            }) : () -> ()
        """,
    )

    print_as_mlir_and_compare(
        """module() {
           %0 : !param_attr = any()
           any(%0 : !param_attr, %0 : !param_attr)
        }""",
        """"module"() ({
              %0 = "any"() : () -> #param_attr
              "any"(%0, %0) : (#param_attr, #param_attr) -> ()
            }) : () -> ()
        """,
    )


def test_op_with_attributes():
    """Test printing an operation with attributes."""
    print_as_mlir_and_compare(
        """any() [ "attr" = !data_attr<42> ]""",
        """"any"() {"attr" = #data_attr<42>} : () -> ()""",
    )


def test_param_custom_format():
    """Test printing an operation with a param attribute with custom format."""
    print_as_mlir_and_compare(
        """any() [ "attr" = !param_custom_format<!param_attr> ]""",
        """"any"() {"attr" = #param_custom_format~~} : () -> ()""",
    )



from __future__ import annotations

import inspect
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from enum import Enum
from frozenlist import FrozenList
from functools import reduce
from inspect import isclass
from typing import (Annotated, Any, Callable, Generic, Sequence, TypeAlias,
                    TypeVar, Union, cast, get_args, get_origin, get_type_hints)
from types import UnionType, GenericAlias

from xdsl.ir import (Attribute, Block, Data, OpResult, Operation,
                     ParametrizedAttribute, Region, SSAValue)
from xdsl.utils.diagnostic import Diagnostic
from xdsl.utils.exceptions import BuilderNotFoundException, VerifyException
from xdsl.utils.hints import is_satisfying_hint

# pyright: reportMissingParameterType=false, reportUnknownParameterType=false


def error(op: Operation, msg: str):
    diag = Diagnostic()
    diag.add_message(op, msg)
    diag.raise_exception(f"{op.name} operation does not verify", op)


class IRDLAnnotations(Enum):
    ParamDefAnnot = 1
    AttributeDefAnnot = 2
    OptAttributeDefAnnot = 3
    SingleBlockRegionAnnot = 4


#   ____                _             _       _
#  / ___|___  _ __  ___| |_ _ __ __ _(_)_ __ | |_ ___
# | |   / _ \| '_ \/ __| __| '__/ _` | | '_ \| __/ __|
# | |__| (_) | | | \__ \ |_| | | (_| | | | | | |_\__ \
#  \____\___/|_| |_|___/\__|_|  \__,_|_|_| |_|\__|___/
#


@dataclass
class AttrConstraint(ABC):
    """Constrain an attribute to a certain value."""

    @abstractmethod
    def verify(self, attr: Attribute) -> None:
        """
        Check if the attribute satisfies the constraint,
        or raise an exception otherwise.
        """
        ...


@dataclass
class EqAttrConstraint(AttrConstraint):
    """Constrain an attribute to be equal to another attribute."""

    attr: Attribute
    """The attribute we want to check equality with."""

    def verify(self, attr: Attribute) -> None:
        if attr != self.attr:
            raise VerifyException(
                f"Expected attribute {self.attr} but got {attr}")


@dataclass
class BaseAttr(AttrConstraint):
    """Constrain an attribute to be of a given base type."""

    attr: type[Attribute]
    """The expected attribute base type."""

    def verify(self, attr: Attribute) -> None:
        if not isinstance(attr, self.attr):
            raise VerifyException(
                f"{attr} should be of base attribute {self.attr.name}")


def attr_constr_coercion(attr: (Attribute | type[Attribute]
                                | AttrConstraint)) -> AttrConstraint:
    """
    Attributes are coerced into EqAttrConstraints,
    and Attribute types are coerced into BaseAttr.
    """
    if isinstance(attr, Attribute):
        return EqAttrConstraint(attr)
    if isclass(attr) and issubclass(attr, Attribute):
        return BaseAttr(attr)
    assert (isinstance(attr, AttrConstraint))
    return attr


@dataclass
class AnyAttr(AttrConstraint):
    """Constraint that is verified by all attributes."""

    def verify(self, attr: Attribute) -> None:
        pass


@dataclass(init=False)
class AnyOf(AttrConstraint):
    """Ensure that an attribute satisfies one of the given constraints."""

    attr_constrs: list[AttrConstraint]
    """The list of constraints that are checked."""

    def __init__(self, attr_constrs: Sequence[Attribute | type[Attribute]
                                              | AttrConstraint]):
        self.attr_constrs = [
            attr_constr_coercion(constr) for constr in attr_constrs
        ]

    def verify(self, attr: Attribute) -> None:
        for attr_constr in self.attr_constrs:
            try:
                attr_constr.verify(attr)
                return
            except VerifyException:
                pass
        raise VerifyException(f"Unexpected attribute {attr}")


@dataclass()
class AllOf(AttrConstraint):
    """Ensure that an attribute satisfies all the given constraints."""

    attr_constrs: list[AttrConstraint]
    """The list of constraints that are checked."""

    def verify(self, attr: Attribute) -> None:
        for attr_constr in self.attr_constrs:
            attr_constr.verify(attr)


@dataclass(init=False)
class ParamAttrConstraint(AttrConstraint):
    """
    Constrain an attribute to be of a given type,
    and also constrain its parameters with additional constraints.
    """

    base_attr: type[ParametrizedAttribute]
    """The base attribute type."""

    param_constrs: list[AttrConstraint]
    """The attribute parameter constraints"""

    def __init__(self, base_attr: type[ParametrizedAttribute],
                 param_constrs: Sequence[(Attribute | type[Attribute]
                                          | AttrConstraint)]):
        self.base_attr = base_attr
        self.param_constrs = [
            attr_constr_coercion(constr) for constr in param_constrs
        ]

    def verify(self, attr: Attribute) -> None:
        if not isinstance(attr, self.base_attr):
            raise VerifyException(
                f"{attr} should be of base attribute {self.base_attr.name}")
        if len(self.param_constrs) != len(attr.parameters):
            raise VerifyException(
                f"{len(self.param_constrs)} parameters expected, "
                f"but got {len(attr.parameters)}")
        for idx, param_constr in enumerate(self.param_constrs):
            param_constr.verify(attr.parameters[idx])


def irdl_to_attr_constraint(
    irdl: Any,
    *,
    allow_type_var: bool = False,
    type_var_mapping: dict[TypeVar, AttrConstraint] | None = None
) -> AttrConstraint:
    if isinstance(irdl, AttrConstraint):
        return irdl

    if isinstance(irdl, Attribute):
        return EqAttrConstraint(irdl)

    # Annotated case
    # Each argument of the Annotated type corresponds to a constraint to satisfy.
    if get_origin(irdl) == Annotated:
        constraints: list[AttrConstraint] = []
        for arg in get_args(irdl):
            # We should not try to convert IRDL annotations, which do not
            # correspond to constraints
            if isinstance(arg, IRDLAnnotations):
                continue
            constraints.append(
                irdl_to_attr_constraint(arg,
                                        allow_type_var=allow_type_var,
                                        type_var_mapping=type_var_mapping))
        if len(constraints) > 1:
            return AllOf(constraints)
        return constraints[0]

    # Attribute class case
    # This is an `AnyAttr`, which does not constrain the attribute.
    if irdl is Attribute:
        return AnyAttr()

    # Attribute class case
    # This is a coercion for an `BaseAttr`.
    if isclass(irdl) and issubclass(irdl, Attribute):
        return BaseAttr(irdl)

    # Type variable case
    # We take the type variable bound constraint.
    if isinstance(irdl, TypeVar):
        if not allow_type_var:
            raise Exception("TypeVar in unexpected context.")
        if type_var_mapping:
            if irdl in type_var_mapping:
                return type_var_mapping[irdl]
        if irdl.__bound__ is None:
            raise Exception("Type variables used in IRDL are expected to"
                            " be bound.")
        # We do not allow nested type variables.
        return irdl_to_attr_constraint(irdl.__bound__)

    origin = get_origin(irdl)

    # GenericData case
    if isclass(origin) and issubclass(origin, GenericData):
        return AllOf([
            BaseAttr(origin),
            origin.generic_constraint_coercion(get_args(irdl))
        ])

    # Generic ParametrizedAttributes case
    # We translate it to constraints over the attribute parameters.
    if isclass(origin) and issubclass(
            origin, ParametrizedAttribute) and issubclass(origin, Generic):
        args = [
            irdl_to_attr_constraint(arg,
                                    allow_type_var=allow_type_var,
                                    type_var_mapping=type_var_mapping)
            for arg in get_args(irdl)
        ]
        generic_args = ()

        # Get the Generic parent class to get the TypeVar parameters
        for parent in origin.__orig_bases__:  # type: ignore
            if get_origin(parent) == Generic:
                generic_args = get_args(parent)
                break
        else:
            raise Exception(
                f"Cannot parametrized non-generic {origin.name} attribute.")

        # Check that we have the right number of parameters
        if len(args) != len(generic_args):
            raise Exception(f"{origin.name} expects {len(generic_args)}"
                            f" parameters, got {len(args)}.")

        type_var_mapping = {
            parameter: arg
            for parameter, arg in zip(generic_args, args)
        }

        origin_parameters = irdl_param_attr_get_param_type_hints(origin)
        origin_constraints = [
            irdl_to_attr_constraint(param,
                                    allow_type_var=True,
                                    type_var_mapping=type_var_mapping)
            for _, param in origin_parameters
        ]
        return ParamAttrConstraint(origin, origin_constraints)

    # Union case
    # This is a coercion for an `AnyOf` constraint.
    if origin == UnionType or origin == Union:
        constraints: list[AttrConstraint] = []
        for arg in get_args(irdl):
            # We should not try to convert IRDL annotations, which do not
            # correspond to constraints
            if isinstance(arg, IRDLAnnotations):
                continue
            constraints.append(
                irdl_to_attr_constraint(arg,
                                        allow_type_var=allow_type_var,
                                        type_var_mapping=type_var_mapping))
        if len(constraints) > 1:
            return AnyOf(constraints)
        return constraints[0]

    # Better error messages for missing GenericData in Data definitions
    if isclass(origin) and issubclass(origin, Data):
        raise ValueError(
            f"Generic `Data` type '{origin.name}' cannot be converted to "
            "an attribute constraint. Consider making it inherit from "
            "`GenericData` instead of `Data`.")

    raise ValueError(f"Unexpected irdl constraint: {irdl}")


#   ___                       _   _
#  / _ \ _ __   ___ _ __ __ _| |_(_) ___  _ __
# | | | | '_ \ / _ \ '__/ _` | __| |/ _ \| '_ \
# | |_| | |_) |  __/ | | (_| | |_| | (_) | | | |
#  \___/| .__/ \___|_|  \__,_|\__|_|\___/|_| |_|
#       |_|

_OpT = TypeVar('_OpT', bound=Operation)


@dataclass
class IRDLOption(ABC):
    """Additional option used in IRDL."""
    ...


@dataclass
class AttrSizedOperandSegments(IRDLOption):
    """Expect an attribute on the op that contains the sizes of the variadic operands."""

    attribute_name = "operand_segment_sizes"
    """Name of the attribute containing the variadic operand sizes."""


@dataclass
class AttrSizedResultSegments(IRDLOption):
    """Expect an attribute on the op that contains the sizes of the variadic results."""

    attribute_name = "result_segment_sizes"
    """Name of the attribute containing the variadic result sizes."""


@dataclass
class OperandOrResultDef(ABC):
    """An operand or a result definition. Should not be used directly."""
    ...


@dataclass
class VariadicDef(OperandOrResultDef):
    """A variadic operand or result definition. Should not be used directly."""
    ...


@dataclass
class OptionalDef(VariadicDef):
    """An optional operand or result definition. Should not be used directly."""
    ...


@dataclass(init=False)
class OperandDef(OperandOrResultDef):
    """An IRDL operand definition."""

    constr: AttrConstraint
    """The operand constraint."""

    def __init__(self, typ: Attribute | type[Attribute] | AttrConstraint):
        self.constr = attr_constr_coercion(typ)


Operand: TypeAlias = SSAValue


@dataclass(init=False)
class VarOperandDef(OperandDef, VariadicDef):
    """An IRDL variadic operand definition."""


VarOperand: TypeAlias = list[SSAValue]


@dataclass(init=False)
class OptOperandDef(VarOperandDef, OptionalDef):
    """An IRDL optional operand definition."""


OptOperand: TypeAlias = SSAValue | None


@dataclass(init=False)
class ResultDef(OperandOrResultDef):
    """An IRDL result definition."""

    constr: AttrConstraint
    """The result constraint."""

    def __init__(self, typ: Attribute | type[Attribute] | AttrConstraint):
        self.constr = attr_constr_coercion(typ)


@dataclass(init=False)
class VarResultDef(ResultDef, VariadicDef):
    """An IRDL variadic result definition."""


VarOpResult: TypeAlias = list[OpResult]


@dataclass(init=False)
class OptResultDef(VarResultDef, OptionalDef):
    """An IRDL optional result definition."""


OptOpResult: TypeAlias = OpResult | None


@dataclass
class RegionDef(Region):
    """
    An IRDL region definition.
    """


@dataclass
class VarRegionDef(RegionDef, VariadicDef):
    """An IRDL variadic region definition."""


@dataclass
class OptRegionDef(RegionDef, OptionalDef):
    """An IRDL optional region definition."""


VarRegion: TypeAlias = list[Region]
OptRegion: TypeAlias = Region | None


@dataclass
class SingleBlockRegionDef(RegionDef):
    """An IRDL region definition that expects exactly one block."""


class VarSingleBlockRegionDef(RegionDef, VariadicDef):
    """An IRDL variadic region definition that expects exactly one block."""


class OptSingleBlockRegionDef(RegionDef, OptionalDef):
    """An IRDL optional region definition that expects exactly one block."""


SingleBlockRegion: TypeAlias = Annotated[
    Region, IRDLAnnotations.SingleBlockRegionAnnot]
VarSingleBlockRegion: TypeAlias = Annotated[
    list[Region], IRDLAnnotations.SingleBlockRegionAnnot]
OptSingleBlockRegion: TypeAlias = Annotated[
    Region | None, IRDLAnnotations.SingleBlockRegionAnnot]


@dataclass(init=False)
class AttributeDef:
    """An IRDL attribute definition."""

    constr: AttrConstraint
    """The attribute constraint."""

    def __init__(self, typ: Attribute | type[Attribute] | AttrConstraint):
        self.constr = attr_constr_coercion(typ)


@dataclass(init=False)
class OptAttributeDef(AttributeDef):
    """An IRDL attribute definition for an optional attribute."""

    def __init__(self, typ: Attribute | type[Attribute] | AttrConstraint):
        super().__init__(typ)


_OpAttrT = TypeVar("_OpAttrT", bound=Attribute)

OpAttr: TypeAlias = Annotated[_OpAttrT, IRDLAnnotations.AttributeDefAnnot]
OptOpAttr: TypeAlias = Annotated[_OpAttrT | None,
                                 IRDLAnnotations.OptAttributeDefAnnot]


@dataclass(kw_only=True)
class OpDef:
    """The internal IRDL definition of an operation."""
    name: str = field(kw_only=False)
    operands: list[tuple[str, OperandDef]] = field(default_factory=list)
    results: list[tuple[str, ResultDef]] = field(default_factory=list)
    attributes: dict[str, AttributeDef] = field(default_factory=dict)
    regions: list[tuple[str, RegionDef]] = field(default_factory=list)
    options: list[IRDLOption] = field(default_factory=list)

    @staticmethod
    def from_pyrdl(pyrdl_def: type[_OpT]) -> OpDef:
        """Decorator used on classes to define a new operation definition."""

        # Get all fields of the class, including the parent classes
        clsdict: dict[str, Any] = dict()
        for parent_cls in pyrdl_def.mro()[::-1]:
            clsdict = {**clsdict, **parent_cls.__dict__}

        if "name" not in clsdict:
            raise Exception(
                f"pyrdl operation definition '{pyrdl_def.__name__}' does not "
                "define the operation name. The operation name is defined by "
                "adding a 'name' field.")

        op_def = OpDef(clsdict["name"])
        for field_name, field_type in get_type_hints(
                pyrdl_def, include_extras=True).items():

            if field_name in get_type_hints(Operation).keys():
                continue

            # If the field type is an Annotated, separate the origin
            # from the arguments.
            # If the field type is not an Annotated, then the arguments should
            # just be the field itself.
            origin = get_origin(field_type)
            args: tuple[Any, ...]
            if origin is None:
                args = (field_type, )
            elif origin == Annotated:
                args = get_args(field_type)
            else:
                args = (field_type, )
            args = cast(tuple[Any, ...], args)

            # Get attribute constraints from a list of pyrdl constraints
            def get_constraint(
                    pyrdl_constrs: tuple[Any, ...]) -> AttrConstraint:
                constraints = [
                    irdl_to_attr_constraint(pyrdl_constr)
                    for pyrdl_constr in pyrdl_constrs
                    if not isinstance(pyrdl_constr, IRDLAnnotations)
                ]
                if len(constraints) == 0:
                    return AnyAttr()
                if len(constraints) == 1:
                    return constraints[0]
                return AllOf(constraints)

            # Get the operand, result, attribute, or region definition, from
            # the pyrdl description.

            # For operands and results, constrants are encoded as arguments of
            # an Annotated, where the origin is the definition type (operand,
            # optional result, etc...).
            # For Attributes, constraints are encoded in the origin and the
            # args of the Annotated, and the definition type (required or
            # optional) is given in the Annotated arguments.
            # For Regions, SingleBlock regions are given as Annotated arguments,
            # and otherwise the Annotated origin (if it is an Annotated) gives
            # the Region definition (required, optional, or variadic).

            # Operand annotation
            if args[0] == Operand:
                constraint = get_constraint(args[1:])
                op_def.operands.append((field_name, OperandDef(constraint)))
            elif args[0] == list[Operand]:
                constraint = get_constraint(args[1:])
                op_def.operands.append((field_name, VarOperandDef(constraint)))
            elif args[0] == (Operand | None):
                constraint = get_constraint(args[1:])
                op_def.operands.append((field_name, OptOperandDef(constraint)))

            # Result annotation
            elif args[0] == OpResult:
                constraint = get_constraint(args[1:])
                op_def.results.append((field_name, ResultDef(constraint)))
            elif args[0] == list[OpResult]:
                constraint = get_constraint(args[1:])
                op_def.results.append((field_name, VarResultDef(constraint)))
            elif args[0] == (OpResult | None):
                constraint = get_constraint(args[1:])
                op_def.results.append((field_name, OptResultDef(constraint)))

            # Attribute annotation
            elif IRDLAnnotations.AttributeDefAnnot in args:
                constraint = get_constraint(args)
                op_def.attributes[field_name] = AttributeDef(constraint)
            elif IRDLAnnotations.OptAttributeDefAnnot in args:
                assert get_origin(args[0]) in [UnionType, Union]
                args = (reduce(lambda x, y: x | y,
                               get_args(args[0])[:-1]), *args[1:])
                constraint = get_constraint(args)
                op_def.attributes[field_name] = OptAttributeDef(constraint)

            # Region annotation
            elif args[0] == Region:
                if (len(args) > 1
                        and args[1] == IRDLAnnotations.SingleBlockRegionAnnot):
                    op_def.regions.append((field_name, SingleBlockRegionDef()))
                else:
                    op_def.regions.append((field_name, RegionDef()))
            elif args[0] == VarRegion:
                if (len(args) > 1
                        and args[1] == IRDLAnnotations.SingleBlockRegionAnnot):
                    op_def.regions.append(
                        (field_name, VarSingleBlockRegionDef()))
                else:
                    op_def.regions.append((field_name, VarRegionDef()))
            elif args[0] == OptRegion:
                if (len(args) > 1
                        and args[1] == IRDLAnnotations.SingleBlockRegionAnnot):
                    op_def.regions.append(
                        (field_name, OptSingleBlockRegionDef()))
                else:
                    op_def.regions.append((field_name, OptRegionDef()))

        op_def.options = clsdict.get("irdl_options", [])

        return op_def


class VarIRConstruct(Enum):
    """
    An enum representing the part of an IR that may be variadic.
    This contains operands, results, and regions.
    """
    OPERAND = 1
    RESULT = 2
    REGION = 3


def get_construct_name(construct: VarIRConstruct) -> str:
    """Get the type name, this is used mostly for error messages."""
    if construct == VarIRConstruct.OPERAND:
        return "operand"
    if construct == VarIRConstruct.RESULT:
        return "result"
    if construct == VarIRConstruct.REGION:
        return "region"
    assert False, "Unknown VarIRConstruct value"


def get_construct_defs(
    op_def: OpDef, construct: VarIRConstruct
) -> list[tuple[str, OperandDef]] | list[tuple[str, ResultDef]] | list[tuple[
        str, RegionDef]]:
    """Get the definitions of this type in an operation definition."""
    if construct == VarIRConstruct.OPERAND:
        return op_def.operands
    if construct == VarIRConstruct.RESULT:
        return op_def.results
    if construct == VarIRConstruct.REGION:
        return op_def.regions
    assert False, "Unknown VarIRConstruct value"


def get_op_constructs(
    op: Operation, construct: VarIRConstruct
) -> FrozenList[SSAValue] | list[OpResult] | list[Region]:
    """
        Get the list of arguments of the type in an operation.
        For example, if the argument type is an operand, get the list of
        operands.
        """
    if construct == VarIRConstruct.OPERAND:
        return op.operands
    if construct == VarIRConstruct.RESULT:
        return op.results
    if construct == VarIRConstruct.REGION:
        return op.regions
    assert False, "Unknown VarIRConstruct value"


def get_attr_size_option(
    construct: VarIRConstruct
) -> AttrSizedOperandSegments | AttrSizedResultSegments | None:
    """Get the AttrSized option for this type."""
    if construct == VarIRConstruct.OPERAND:
        return AttrSizedOperandSegments()
    if construct == VarIRConstruct.RESULT:
        return AttrSizedResultSegments()
    if construct == VarIRConstruct.REGION:
        return None
    assert False, "Unknown VarIRConstruct value"


def get_variadic_sizes_from_attr(op: Operation,
                                 defs: Sequence[tuple[str,
                                                      OperandDef | ResultDef]],
                                 construct: VarIRConstruct,
                                 size_attribute_name: str) -> list[int]:
    """
    Get the sizes of the variadic definitions
    from the corresponding attribute.
    """
    # Circular import because DenseIntOrFPElementsAttr is defined using IRDL
    from xdsl.dialects.builtin import DenseIntOrFPElementsAttr

    # Check that the attribute is present
    if size_attribute_name not in op.attributes:
        raise VerifyException(
            f"Expected {size_attribute_name} attribute in {op.name} operation."
        )
    attribute = op.attributes[size_attribute_name]
    if not isinstance(attribute, DenseIntOrFPElementsAttr):
        raise VerifyException(f"{size_attribute_name} attribute is expected "
                              "to be a DenseIntOrFPElementsAttr.")
    def_sizes: list[int] = [
        size_attr.value.data for size_attr in attribute.data.data
    ]
    if len(def_sizes) != len(defs):
        raise VerifyException(
            f"expected {len(defs)} values in "
            f"{size_attribute_name}, but got {len(def_sizes)}")

    variadic_sizes = list[int]()
    for ((arg_name, arg_def), arg_size) in zip(defs, def_sizes):
        if isinstance(arg_def, OptionalDef) and arg_size > 1:
            raise VerifyException(
                f"optional {get_construct_name(construct)} {arg_name} is expected to "
                f"be of size 0 or 1 in {size_attribute_name}, but got "
                f"{arg_size}")

        if not isinstance(arg_def, VariadicDef) and arg_size != 1:
            raise VerifyException(
                f"non-variadic {get_construct_name(construct)} {arg_name} is expected "
                f"to be of size 0 or 1 in {size_attribute_name}, but got "
                f"{arg_size}")

        if isinstance(arg_def, VariadicDef):
            variadic_sizes.append(arg_size)

    return variadic_sizes


def get_variadic_sizes(op: Operation, op_def: OpDef,
                       construct: VarIRConstruct) -> list[int]:
    """Get variadic sizes of operands or results."""

    defs = get_construct_defs(op_def, construct)
    args = get_op_constructs(op, construct)
    def_type_name = get_construct_name(construct)
    attribute_option = get_attr_size_option(construct)

    variadic_defs = [(arg_name, arg_def) for arg_name, arg_def in defs
                     if isinstance(arg_def, VariadicDef)]

    # If the size is in the attributes, fetch it
    if (attribute_option is not None) and (attribute_option in op_def.options):
        return get_variadic_sizes_from_attr(op, defs, construct,
                                            attribute_option.attribute_name)

    # If there are no variadics arguments,
    # we just check that we have the right number of arguments
    if len(variadic_defs) == 0:
        if len(args) != len(defs):
            raise VerifyException(
                f"Expected {len(defs)} {def_type_name}, but got {len(args)}")
        return []

    # If there is a single variadic argument,
    # we can get its size from the number of arguments.
    if len(variadic_defs) == 1:
        if len(args) - len(defs) + 1 < 0:
            raise VerifyException(f"Expected at least {len(defs) - 1} "
                                  f"{def_type_name}s, got {len(defs)}")
        return [len(args) - len(defs) + 1]

    # Unreachable, all cases should have been handled.
    # Additional cases should raise an exception upon
    # definition of the irdl operation.
    assert False, "Unexpected xDSL error while fetching variadic sizes"


def get_operand_result_or_region(
    op: Operation, op_def: OpDef, arg_def_idx: int, previous_var_args: int,
    construct: VarIRConstruct
) -> None | SSAValue | FrozenList[SSAValue] | list[OpResult] | Region | list[
        Region]:
    """
    Get an operand, result, or region.
    In the case of a variadic definition, return a list of elements.
    :param op: The operation we want to get argument of.
    :param arg_def_idx: The index of the argument in the irdl definition.
    :param previous_var_args: The number of previous variadic definitions
           before this definition.
    :param arg_type: The type of the argument we want
           (i.e. operand, result, or region)
    :return:
    """
    defs = get_construct_defs(op_def, construct)
    args = get_op_constructs(op, construct)

    variadic_sizes = get_variadic_sizes(op, op_def, construct)

    begin_arg = arg_def_idx - previous_var_args + sum(
        variadic_sizes[:previous_var_args])
    if isinstance(defs[arg_def_idx][1], OptionalDef):
        arg_size = variadic_sizes[previous_var_args]
        if arg_size == 0:
            return None
        else:
            return args[begin_arg]
    if isinstance(defs[arg_def_idx][1], VariadicDef):
        arg_size = variadic_sizes[previous_var_args]
        return args[begin_arg:begin_arg + arg_size]
    else:
        return args[begin_arg]


def irdl_op_verify_arg_list(op: Operation, op_def: OpDef,
                            construct: VarIRConstruct) -> None:
    """Verify the argument list of an operation."""
    arg_sizes = get_variadic_sizes(op, op_def, construct)
    arg_idx = 0
    var_idx = 0
    args = get_op_constructs(op, construct)

    def verify_arg(arg: Any, arg_def: Any, arg_idx: int) -> None:
        """Verify a single argument."""
        try:
            if construct == VarIRConstruct.OPERAND or construct == VarIRConstruct.RESULT:
                arg_def.constr.verify(arg.typ)
            elif construct == VarIRConstruct.REGION:
                if isinstance(arg_def,
                              SingleBlockRegionDef) and len(arg.blocks) != 1:
                    raise VerifyException("expected a single block, but got "
                                          f"{len(arg.blocks)} blocks")
            else:
                assert False, "Unknown ArgType value"
        except Exception as e:
            error(
                op, f"{get_construct_name(construct)} at position "
                f"{arg_idx} does not verify!\n{e}")

    for def_idx, (_,
                  arg_def) in enumerate(get_construct_defs(op_def, construct)):
        if isinstance(arg_def, VariadicDef):
            for _ in range(arg_sizes[var_idx]):
                verify_arg(args[arg_idx], arg_def, def_idx)
                arg_idx += 1
            var_idx += 1
        else:
            verify_arg(args[arg_idx], arg_def, def_idx)
            arg_idx += 1


def irdl_op_verify(op: Operation, op_def: OpDef) -> None:
    """Given an IRDL definition, verify that an operation satisfies its invariants."""

    # Verify operands.
    irdl_op_verify_arg_list(op, op_def, VarIRConstruct.OPERAND)

    # Verify results.
    irdl_op_verify_arg_list(op, op_def, VarIRConstruct.RESULT)

    # Verify regions.
    irdl_op_verify_arg_list(op, op_def, VarIRConstruct.REGION)

    for attr_name, attr_def in op_def.attributes.items():
        if attr_name not in op.attributes:
            if isinstance(attr_def, OptAttributeDef):
                continue
            raise VerifyException(f"attribute {attr_name} expected")
        attr_def.constr.verify(op.attributes[attr_name])


def irdl_build_attribute(irdl_def: AttrConstraint, result: Any) -> Attribute:
    if isinstance(irdl_def, BaseAttr):
        if isinstance(result, tuple):
            return irdl_def.attr.build(*result)
        return irdl_def.attr.build(result)
    if isinstance(result, Attribute):
        return result
    raise Exception(f"builder expected an attribute, got {result}")


def irdl_build_arg_list(construct: VarIRConstruct,
                        args: Sequence[Any],
                        arg_defs: Sequence[tuple[str, Any]],
                        error_prefix: str = "") -> tuple[list[Any], list[int]]:
    """Build a list of arguments (operands, results, regions)"""

    def build_arg(arg_def: Any, arg: Any) -> Any:
        """Build a single argument."""
        if construct == VarIRConstruct.OPERAND:
            return SSAValue.get(arg)
        elif construct == VarIRConstruct.RESULT:
            assert isinstance(arg_def, ResultDef)
            return irdl_build_attribute(arg_def.constr, arg)
        elif construct == VarIRConstruct.REGION:
            assert isinstance(arg_def, RegionDef)
            return Region.get(arg)
        else:
            assert False, "Unknown ArgType value"

    if len(args) != len(arg_defs):
        raise ValueError(
            f"Expected {len(arg_defs)} {get_construct_name(construct)}, "
            f"but got {len(args)}")

    res = list[Any]()
    arg_sizes = list[int]()

    for arg_idx, ((arg_name, arg_def), arg) in enumerate(zip(arg_defs, args)):
        if isinstance(arg_def, VariadicDef):
            if not isinstance(arg, list):
                raise ValueError(
                    error_prefix +
                    f"variadic {construct} {arg_idx} '{arg_name}' "
                    f" expects a list, but got {arg}")
            arg = cast(list[Any], arg)

            # Check we have at most one argument for optional defintions.
            if isinstance(arg_def, OptionalDef) and len(arg) > 1:
                raise ValueError(
                    error_prefix +
                    f"optional {construct} {arg_idx} '{arg_name}' "
                    "expects a list of size at most 1, but "
                    f"got a list of size {len(arg)}")

            res.extend([build_arg(arg_def, arg_arg) for arg_arg in arg])
            arg_sizes.append(len(arg))
        else:
            res.append(build_arg(arg_def, arg))
            arg_sizes.append(1)
    return res, arg_sizes


def irdl_op_builder(cls: type[_OpT], op_def: OpDef,
                    operands: Sequence[SSAValue | Operation
                                       | list[SSAValue | Operation] | None],
                    res_types: Sequence[Any | list[Any] | None],
                    attributes: dict[str, Any], successors: Sequence[Block],
                    regions: Sequence[Any | None]) -> _OpT:
    """Builder for an irdl operation."""

    # We need irdl to define DenseIntOrFPElementsAttr, but here we need
    # DenseIntOrFPElementsAttr.
    # So we have a circular dependency that we solve by importing in this function.
    from xdsl.dialects.builtin import (DenseIntOrFPElementsAttr, i32)

    error_prefix = f"Error in {op_def.name} builder: "

    # Build the operands
    built_operands, operand_sizes = irdl_build_arg_list(
        VarIRConstruct.OPERAND, operands, op_def.operands, error_prefix)

    # Build the results
    built_res_types, result_sizes = irdl_build_arg_list(
        VarIRConstruct.RESULT, res_types, op_def.results, error_prefix)

    # Build the regions
    built_regions, _ = irdl_build_arg_list(VarIRConstruct.REGION, regions,
                                           op_def.regions, error_prefix)

    # Build attributes by forwarding the values to the attribute builders
    attr_defs = {name: def_ for (name, def_) in op_def.attributes.items()}

    built_attributes = dict[str, Attribute]()
    for attr_name, attr in attributes.items():
        if attr_name not in attr_defs:
            if isinstance(attr, Attribute):
                built_attributes[attr_name] = attr
                continue
            raise ValueError(error_prefix +
                             f"unexpected attribute name {attr_name}.")
        built_attributes[attr_name] = irdl_build_attribute(
            attr_defs[attr_name].constr, attr)

    # Take care of variadic operand and result segment sizes.
    if AttrSizedOperandSegments() in op_def.options:
        sizes = operand_sizes
        built_attributes[AttrSizedOperandSegments.attribute_name] =\
            DenseIntOrFPElementsAttr.vector_from_list(sizes, i32)

    if AttrSizedResultSegments() in op_def.options:
        sizes = result_sizes
        built_attributes[AttrSizedResultSegments.attribute_name] =\
            DenseIntOrFPElementsAttr.vector_from_list(sizes, i32)

    return cls.create(operands=built_operands,
                      result_types=built_res_types,
                      attributes=built_attributes,
                      successors=successors,
                      regions=built_regions)


def irdl_op_arg_definition(new_attrs: dict[str, Any],
                           construct: VarIRConstruct, op_def: OpDef) -> None:
    previous_variadics = 0
    defs = get_construct_defs(op_def, construct)
    for arg_idx, (arg_name, arg_def) in enumerate(defs):

        def fun(self: Any,
                idx: int = arg_idx,
                previous_vars: int = previous_variadics):
            return get_operand_result_or_region(self, op_def, idx,
                                                previous_vars, construct)

        new_attrs[arg_name] = property(fun)
        if isinstance(arg_def, VariadicDef):
            previous_variadics += 1

    # If we have multiple variadics, check that we have an
    # attribute that holds the variadic sizes.
    arg_size_option = get_attr_size_option(construct)
    if previous_variadics > 1 and (arg_size_option is None
                                   or arg_size_option not in op_def.options):
        if arg_size_option is None:
            arg_size_option_name = 'unknown'
        else:
            arg_size_option_name = arg_size_option.__name__  # type: ignore
        raise Exception(
            "Operation defines more than two variadic "
            f"{get_construct_name(construct)}s, but do not define the "
            f"{arg_size_option_name} PyRDL option.")


def irdl_op_definition(cls: type[_OpT]) -> type[_OpT]:
    """Decorator used on classes to define a new operation definition."""

    assert issubclass(
        cls,
        Operation), f"class {cls.__name__} should be a subclass of Operation"

    # Get all fields of the class, including the parent classes
    clsdict = dict[str, Any]()
    for parent_cls in cls.mro()[::-1]:
        clsdict = {**clsdict, **parent_cls.__dict__}

    op_def = OpDef.from_pyrdl(cls)
    new_attrs = dict[str, Any]()

    # Add operand access fields
    irdl_op_arg_definition(new_attrs, VarIRConstruct.OPERAND, op_def)

    # Add result access fields
    irdl_op_arg_definition(new_attrs, VarIRConstruct.RESULT, op_def)

    # Add region access fields
    irdl_op_arg_definition(new_attrs, VarIRConstruct.REGION, op_def)

    for attribute_name, attr_def in op_def.attributes.items():
        if isinstance(attr_def, OptAttributeDef):
            new_attrs[attribute_name] = property(
                lambda self, name=attribute_name: self.attributes.get(
                    name, None))
        else:
            new_attrs[attribute_name] = property(
                lambda self, name=attribute_name: self.attributes[name])

    new_attrs["verify_"] = lambda op: irdl_op_verify(op, op_def)
    if "verify_" in clsdict:
        custom_verifier = clsdict["verify_"]

        def new_verifier(verifier, op):
            verifier(op)
            custom_verifier(op)

        new_attrs["verify_"] = (
            lambda verifier: lambda op: new_verifier(verifier, op))(
                new_attrs["verify_"])

    def builder(cls,
                operands=[],
                result_types=[],
                attributes=dict(),
                successors=[],
                regions=[]):
        return irdl_op_builder(cls, op_def, operands, result_types, attributes,
                               successors, regions)

    new_attrs["build"] = classmethod(builder)
    new_attrs["irdl_definition"] = classmethod(property(lambda cls: op_def))

    return type(cls.__name__, cls.__mro__, {**cls.__dict__, **new_attrs})


#     _   _   _        _ _           _
#    / \ | |_| |_ _ __(_) |__  _   _| |_ ___
#   / _ \| __| __| '__| | '_ \| | | | __/ _ \
#  / ___ \ |_| |_| |  | | |_) | |_| | ||  __/
# /_/   \_\__|\__|_|  |_|_.__/ \__,_|\__\___|
#

_AttrT = TypeVar('_AttrT', bound=Attribute)

_BuilderTyT = TypeVar("_BuilderTyT", bound=Attribute)

BuilderTy: TypeAlias = Callable[..., _BuilderTyT]

IRDL_IS_BUILDER = '__irdl_is_builder'


def builder(f: BuilderTy[_AttrT]) -> BuilderTy[_AttrT]:
    """
    Annotate a function and mark it as an IRDL builder.
    This should only be used as decorator in classes decorated by irdl_attr_builder.
    """
    setattr(f, IRDL_IS_BUILDER, True)
    return f


def irdl_get_builders(cls: type[_AttrT]) -> list[BuilderTy[_AttrT]]:
    """Get functions decorated with 'builder' in a class."""
    builders = list[BuilderTy[_AttrT]]()
    for field_name in cls.__dict__:
        field_ = cls.__dict__[field_name]
        # Builders are staticmethods, so we need to get back the original function
        # with __func__
        if hasattr(field_, "__func__") and hasattr(field_.__func__,
                                                   IRDL_IS_BUILDER):
            builders.append(field_.__func__)
    return builders


#  ____        _
# |  _ \  __ _| |_ __ _
# | | | |/ _` | __/ _` |
# | |_| | (_| | || (_| |
# |____/ \__,_|\__\__,_|
#

_DataElement = TypeVar("_DataElement")


@dataclass(frozen=True)
class GenericData(Data[_DataElement], ABC):
    """
    A Data with type parameters.
    """

    @staticmethod
    @abstractmethod
    def generic_constraint_coercion(args: tuple[Any]) -> AttrConstraint:
        """
        Given the generic parameters passed to the generic attribute type,
        return the corresponding attribute constraint.
        """


_DT = TypeVar("_DT")


def irdl_data_verify(data: Data[_DT], typ: type[_DT]) -> None:
    """Check that the Data has the expected type."""
    if isinstance(data.data, typ):
        return
    raise VerifyException(
        f"{data.name} data attribute expected type {typ}, but {type(data.data)} given."
    )


T = TypeVar('T', bound=Data[Any])


def irdl_data_definition(cls: type[T]) -> type[T]:
    """Decorator to transform an IRDL Data definition to a Python class."""
    new_attrs = dict[str, Any]()

    # Build method is added for all definitions.
    if "build" in cls.__dict__:
        raise Exception(
            f'"build" method for {cls.__name__} is reserved for IRDL, '
            f'and should not be defined.')
    builders = irdl_get_builders(cls)
    new_attrs["build"] = lambda *args: irdl_attr_builder(cls, builders, *args)

    # Verify method is added if not redefined by the user.
    if "verify" not in cls.__dict__:
        for parent in cls.__orig_bases__:
            if get_origin(parent) != Data:
                continue
            if len(get_args(parent)) != 1:
                raise Exception(f"In {cls.__name__} definition: Data expects "
                                "a single type parameter")
            expected_type = get_args(parent)[0]
            if not isclass(expected_type):
                raise Exception(f'In {cls.__name__} definition: Cannot infer '
                                f'"verify" method. Type parameter of Data is '
                                f'not a class.')
            if isinstance(expected_type, GenericAlias):
                raise Exception(f'In {cls.__name__} definition: Cannot infer '
                                f'"verify" method. Type parameter of Data has '
                                f'type GenericAlias.')
            new_attrs[
                "verify"] = lambda self, expected_type=expected_type: irdl_data_verify(
                    self, expected_type)
            break
        else:
            raise Exception(f'Missing method "verify" in {cls.__name__} data '
                            'attribute definition: the "verify" method cannot '
                            'be automatically derived for this definition.')

    return dataclass(frozen=True)(type(cls.__name__, (cls, ), {
        **cls.__dict__,
        **new_attrs
    }))


#  ____                              _   _   _
# |  _ \ __ _ _ __ __ _ _ __ ___    / \ | |_| |_ _ __
# | |_) / _` | '__/ _` | '_ ` _ \  / _ \| __| __| '__|
# |  __/ (_| | | | (_| | | | | | |/ ___ \ |_| |_| |
# |_|   \__,_|_|  \__,_|_| |_| |_/_/   \_\__|\__|_|
#

_A = TypeVar("_A", bound=Attribute)

ParameterDef: TypeAlias = Annotated[_A, IRDLAnnotations.ParamDefAnnot]


def irdl_param_attr_get_param_type_hints(
        cls: type[_PAttrT]) -> list[tuple[str, Any]]:
    """Get the type hints of an IRDL parameter definitions."""
    res = list[tuple[str, Any]]()
    for field_name, field_type in get_type_hints(cls,
                                                 include_extras=True).items():
        if field_name == "name" or field_name == "parameters":
            continue

        origin = get_origin(field_type)
        args = get_args(field_type)
        if origin != Annotated or IRDLAnnotations.ParamDefAnnot not in args:
            raise ValueError(
                f"In attribute {cls.__name__} definition: Parameter " +
                f"definition {field_name} should be defined with " +
                f"type `ParameterDef`, got type {field_type}.")

        res.append((field_name, field_type))
    return res


@dataclass
class ParamAttrDef:
    """The IRDL definition of a parametrized attribute."""
    name: str
    parameters: list[tuple[str, AttrConstraint]]

    @staticmethod
    def from_pyrdl(pyrdl_def: type[ParametrizedAttribute]) -> ParamAttrDef:
        # Get the fields from the class and its parents
        clsdict = dict[str, Any]()
        for parent_cls in pyrdl_def.mro()[::-1]:
            clsdict = {**clsdict, **parent_cls.__dict__}

        if "name" not in clsdict:
            raise Exception(
                f"pyrdl attribute definition '{pyrdl_def.__name__}' does not "
                "define the attribute name. The attribute name is defined by "
                "adding a 'name' field.")

        name = clsdict["name"]

        param_hints = irdl_param_attr_get_param_type_hints(pyrdl_def)

        parameters = list[tuple[str, AttrConstraint]]()
        for param_name, param_type in param_hints:
            constraint = irdl_to_attr_constraint(param_type,
                                                 allow_type_var=True)
            parameters.append((param_name, constraint))

        return ParamAttrDef(name, parameters)


def irdl_attr_verify(attr: ParametrizedAttribute, attr_def: ParamAttrDef):
    """Given an IRDL definition, verify that an attribute satisfies its invariants."""

    if len(attr.parameters) != len(attr_def.parameters):
        raise VerifyException(
            f"In {attr_def.name} attribute verifier: "
            f"{len(attr_def.parameters)} parameters expected, got "
            f"{len(attr.parameters)}")

    for param, (_, param_def) in zip(attr.parameters, attr_def.parameters):
        param_def.verify(param)


_PAttrT = TypeVar('_PAttrT', bound=ParametrizedAttribute)


def irdl_attr_try_builder(
        builder: BuilderTy[_PAttrT],
        *args: tuple[Any, ...]) -> ParametrizedAttribute | None:
    params_dict = get_type_hints(builder)
    builder_params = inspect.signature(builder).parameters
    params = [params_dict[param.name] for param in builder_params.values()]
    defaults = [param.default for param in builder_params.values()]
    num_non_defaults = defaults.count(inspect.Signature.empty)
    if num_non_defaults > len(args):
        return None
    if len(params) < len(args):
        return None
    for arg, param in zip(args, params[:len(args)]):
        if not is_satisfying_hint(arg, param):
            return None
    return builder(*args, *defaults[len(args):])


def irdl_attr_builder(cls: type[_PAttrT],
                      builders: Sequence[BuilderTy[_PAttrT]],
                      *args: tuple[Any, ...]):
    """Try to apply all builders to construct an attribute instance."""
    if len(args) == 1 and isinstance(args[0], cls):
        return args[0]
    for builder in builders:
        res = irdl_attr_try_builder(builder, *args)
        if res:
            return res
    raise BuilderNotFoundException(cls, args)


def irdl_param_attr_definition(cls: type[_PAttrT]) -> type[_PAttrT]:
    """Decorator used on classes to define a new attribute definition."""

    # Get the fields from the class and its parents
    clsdict = dict[str, Any]()
    for parent_cls in cls.mro()[::-1]:
        clsdict = {**clsdict, **parent_cls.__dict__}

    attr_def = ParamAttrDef.from_pyrdl(cls)

    # New fields and methods added to the attribute
    new_fields = dict[str, Any]()

    for idx, (param_name, _) in enumerate(attr_def.parameters):
        new_fields[param_name] = property(
            lambda self, idx=idx: self.parameters[idx])

    new_fields["verify"] = lambda typ: irdl_attr_verify(typ, attr_def)

    if "verify" in clsdict:
        custom_verifier = clsdict["verify"]

        def new_verifier(verifier, op):
            verifier(op)
            custom_verifier(op)

        new_fields["verify"] = (
            lambda verifier: lambda op: new_verifier(verifier, op))(
                new_fields["verify"])

    builders = irdl_get_builders(cls)
    if "build" in cls.__dict__:
        raise Exception(
            f'"build" method for {cls.__name__} is reserved for IRDL, ' +
            'and should not be defined.')
    new_fields["build"] = lambda *args: irdl_attr_builder(cls, builders, *args)

    new_fields["irdl_definition"] = classmethod(property(lambda cls: attr_def))

    return dataclass(frozen=True, init=False)(type(cls.__name__, (cls, ), {
        **cls.__dict__,
        **new_fields
    }))


def irdl_attr_definition(cls: type[_AttrT]) -> type[_AttrT]:
    if issubclass(cls, ParametrizedAttribute):
        return irdl_param_attr_definition(cls)
    if issubclass(cls, Data):
        return irdl_data_definition(cls)
    raise Exception(
        f"Class {cls.__name__} should either be a subclass of 'Data' or "
        "'ParametrizedAttribute'")



"""Custom xDSL exceptions.

This module contains all custom exceptions used by xDSL.
"""
from __future__ import annotations

import sys
import typing
from dataclasses import dataclass
from io import StringIO
from typing import Any

if typing.TYPE_CHECKING:
    from parser import Span, BacktrackingHistory
    from xdsl.ir import Attribute


class DiagnosticException(Exception):
    pass


class VerifyException(DiagnosticException):
    pass


@dataclass
class BuilderNotFoundException(Exception):
    """
    Exception raised when no builders are found for a given attribute type
    and a given tuple of arguments.
    """
    attribute: type[Attribute]
    args: tuple[Any]

    def __str__(self) -> str:
        return f"No builder found for attribute {self.attribute} with " \
               f"arguments {self.args}"


class ParseError(Exception):
    span: 'Span'
    msg: str
    history: 'BacktrackingHistory' | None

    def __init__(self,
                 span: 'Span',
                 msg: str,
                 history: 'BacktrackingHistory' | None = None):
        preamble = ""
        if history:
            preamble = history.error.args[0] + '\n'
        if span is None:
            raise ValueError("Span can't be None!")
        super().__init__(preamble + span.print_with_context(msg))
        self.span = span
        self.msg = msg
        self.history = history

    def print_pretty(self, file=sys.stderr):
        print(self.span.print_with_context(self.msg), file=file)

    def print_with_history(self, file=sys.stderr):
        if self.history is not None:
            for h in sorted(self.history.iterate(), key=lambda h: -h.pos):
                h.print()
        else:
            self.print_pretty(file)

    def __repr__(self):
        io = StringIO()
        self.print_with_history(io)
        return "{}:\n{}".format(self.__class__.__name__, io.getvalue())


class MultipleSpansParseError(ParseError):
    ref_text: str | None
    refs: list[tuple['Span', str]]

    def __init__(
        self,
        span: 'Span',
        msg: str,
        ref_text: str,
        refs: list[tuple['Span', str | None]],
        history: 'BacktrackingHistory' | None = None,
    ):
        super(MultipleSpansParseError, self).__init__(span, msg, history)
        self.refs = refs
        self.ref_text = ref_text

    def print_pretty(self, file=sys.stderr):
        super(MultipleSpansParseError, self).print_pretty(file)
        print(self.ref_text or "With respect to:", file=file)
        for span, msg in self.refs:
            print(span.print_with_context(msg), file=file)



from __future__ import annotations
import inspect
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Callable

from xdsl.dialects.builtin import ModuleOp
from xdsl.ir import (Operation, OpResult, Region, Block, BlockArgument,
                     Attribute, SSAValue)
from xdsl.rewriter import Rewriter


@dataclass(eq=False)
class PatternRewriter:
    """
    A rewriter used during pattern matching.
    Once an operation is matched, this rewriter is used to apply
    modification to the operation and its children.
    """
    current_operation: Operation
    """The matched operation."""

    has_erased_matched_operation: bool = field(default=False, init=False)
    """Was the matched operation erased."""

    added_operations_before: list[Operation] = field(default_factory=list,
                                                     init=False)
    """The operations added directly before the matched operation."""

    added_operations_after: list[Operation] = field(default_factory=list,
                                                    init=False)
    """The operations added directly after the matched operation."""

    has_done_action: bool = field(default=False, init=False)
    """Has the rewriter done any action during the current match."""

    def _can_modify_op(self, op: Operation) -> bool:
        """Check if the operation and its children can be modified by this rewriter."""
        if op == self.current_operation:
            return True
        if op.parent is None:
            return self.current_operation.get_toplevel_object() is not op
        return self._can_modify_block(op.parent)

    def _can_modify_block(self, block: Block) -> bool:
        """Check if the block and its children can be modified by this rewriter."""
        if block.parent is None:
            return True  # Toplevel operation of current_operation is always a ModuleOp
        return self._can_modify_region(block.parent)

    def _can_modify_region(self, region: Region) -> bool:
        """Check if the region and its children can be modified by this rewriter."""
        if region.parent is None:
            return True  # Toplevel operation of current_operation is always a ModuleOp
        return self._can_modify_op(region.parent)

    def insert_op_before_matched_op(self, op: (Operation | list[Operation])):
        """Insert operations before the matched operation."""
        if self.current_operation.parent is None:
            raise Exception(
                "Cannot insert an operation before a toplevel operation.")
        self.has_done_action = True
        block = self.current_operation.parent
        op = op if isinstance(op, list) else [op]
        if len(op) == 0:
            return
        op_idx = block.get_operation_index(self.current_operation)
        block.insert_op(op, op_idx)
        self.added_operations_before += op

    def insert_op_after_matched_op(self, op: (Operation | list[Operation])):
        """Insert operations after the matched operation."""
        if self.current_operation.parent is None:
            raise Exception(
                "Cannot insert an operation after a toplevel operation.")
        self.has_done_action = True
        block = self.current_operation.parent
        op = op if isinstance(op, list) else [op]
        if len(op) == 0:
            return
        op_idx = block.get_operation_index(self.current_operation)
        block.insert_op(op, op_idx + 1)
        self.added_operations_after += op

    def insert_op_at_pos(self, op: Operation | list[Operation], block: Block,
                         pos: int):
        """Insert operations in a block contained in the matched operation."""
        if not self._can_modify_block(block):
            raise Exception("Cannot insert operations in block.")
        self.has_done_action = True
        op = op if isinstance(op, list) else [op]
        if len(op) == 0:
            return
        block.insert_op(op, pos)

    def insert_op_before(self, op: Operation | list[Operation],
                         target_op: Operation):
        """Insert operations before an operation contained in the matched operation."""
        if target_op.parent is None:
            raise Exception(
                "Cannot insert operations before toplevel operation.")
        target_block = target_op.parent
        if not self._can_modify_block(target_block):
            raise Exception("Cannot insert operations in this block.")
        self.has_done_action = True
        op = op if isinstance(op, list) else [op]
        if len(op) == 0:
            return
        target_op.parent.insert_op(op,
                                   target_block.get_operation_index(target_op))

    def insert_op_after(self, op: Operation | list[Operation],
                        target_op: Operation):
        """Insert operations after an operation contained in the matched operation."""
        if target_op.parent is None:
            raise Exception(
                "Cannot insert operations after toplevel operation.")
        target_block = target_op.parent
        if not self._can_modify_block(target_block):
            raise Exception("Cannot insert operations in this block.")
        self.has_done_action = True
        op = op if isinstance(op, list) else [op]
        if len(op) == 0:
            return
        target_op.parent.insert_op(
            op,
            target_block.get_operation_index(target_op) + 1)

    def erase_matched_op(self, safe_erase: bool = True):
        """
        Erase the operation that was matched to.
        If safe_erase is True, check that the operation has no uses.
        Otherwise, replace its uses with ErasedSSAValue.
        """
        self.has_done_action = True
        self.has_erased_matched_operation = True
        Rewriter.erase_op(self.current_operation, safe_erase=safe_erase)

    def erase_op(self, op: Operation, safe_erase: bool = True):
        """
        Erase an operation contained in the matched operation children.
        If safe_erase is True, check that the operation has no uses.
        Otherwise, replace its uses with ErasedSSAValue.
        """
        self.has_done_action = True
        if op == self.current_operation:
            return self.erase_matched_op()
        if not self._can_modify_op(op):
            raise Exception(
                "PatternRewriter can only erase operations that are the matched operation"
                ", or that are contained in the matched operation.")
        Rewriter.erase_op(op, safe_erase=safe_erase)

    def replace_matched_op(self,
                           new_ops: Operation | list[Operation],
                           new_results: list[SSAValue | None] | None = None,
                           safe_erase: bool = True):
        """
        Replace the matched operation with new operations.
        Also, optionally specify SSA values to replace the operation results.
        If safe_erase is True, check that the operation has no uses.
        Otherwise, replace its uses with ErasedSSAValue.
        """
        self.has_done_action = True
        if not isinstance(new_ops, list):
            new_ops = [new_ops]
        self.has_erased_matched_operation = True
        Rewriter.replace_op(self.current_operation,
                            new_ops,
                            new_results,
                            safe_erase=safe_erase)
        self.added_operations_before += new_ops

    def replace_op(self,
                   op: Operation,
                   new_ops: Operation | list[Operation],
                   new_results: list[SSAValue | None] | None = None,
                   safe_erase: bool = True):
        """
        Replace an operation with new operations.
        The operation should be a child of the matched operation.
        Also, optionally specify SSA values to replace the operation results.
        If safe_erase is True, check that the operation has no uses.
        Otherwise, replace its uses with ErasedSSAValue.
        """
        self.has_done_action = True
        if op == self.current_operation:
            return self.replace_matched_op(new_ops, new_results, safe_erase)
        if not self._can_modify_op(op):
            raise Exception(
                "PatternRewriter can only replace operations that are the matched "
                "operation, or that are contained in the matched operation.")
        Rewriter.replace_op(op, new_ops, new_results, safe_erase=safe_erase)

    def modify_block_argument_type(self, arg: BlockArgument,
                                   new_type: Attribute):
        """
        Modify the type of a block argument.
        The block should be contained in the matched operation.
        """
        if not self._can_modify_block(arg.block):
            raise Exception(
                "Cannot modify blocks that are not contained in the matched operation"
            )
        self.has_done_action = True
        arg.typ = new_type

    def insert_block_argument(self, block: Block, index: int,
                              typ: Attribute) -> BlockArgument:
        """
        Insert a new block argument.
        The block should be contained in the matched operation.
        """
        if not self._can_modify_block(block):
            raise Exception(
                "Cannot modify blocks that are not contained in the matched operation"
            )
        self.has_done_action = True
        return block.insert_arg(typ, index)

    def erase_block_argument(self,
                             arg: BlockArgument,
                             safe_erase: bool = True) -> None:
        """
        Erase a new block argument.
        The block should be contained in the matched operation.
        If safe_erase is true, then raise an exception if the block argument has still
        uses, otherwise, replace it with an ErasedSSAValue.
        """
        if not self._can_modify_block(arg.block):
            raise Exception(
                "Cannot modify blocks that are not contained in the matched operation"
            )
        self.has_done_action = True
        arg.block.erase_arg(arg, safe_erase=safe_erase)

    def inline_block_at_pos(self, block: Block, target_block: Block, pos: int):
        """
        Move the block operations to a given position in another block.
        This block should not be a parent of the block to move to, and both blocks
        should be child of the matched operation.
        """
        self.has_done_action = True
        if not self._can_modify_block(
                target_block) or not self._can_modify_block(block):
            raise Exception(
                "Cannot modify blocks that are not contained in the matched operation."
            )
        Rewriter.inline_block_at_pos(block, target_block, pos)

    def inline_block_before_matched_op(self, block: Block):
        """
        Move the block operations before the matched operation.
        The block should not be a parent of the operation, and should be a child of the
        matched operation.
        """
        self.has_done_action = True
        if not self._can_modify_block(block):
            raise Exception(
                "Cannot move blocks that are not contained in the matched operation."
            )
        self.added_operations_before += block.ops
        Rewriter.inline_block_before(block, self.current_operation)

    def inline_block_before(self, block: Block, op: Operation):
        """
        Move the block operations before the given operation.
        The block should not be a parent of the operation, and should be a child of the
        matched operation.
        The operation should also be a child of the matched operation.
        """
        self.has_done_action = True
        if op is self.current_operation:
            return self.inline_block_before_matched_op(block)
        if not self._can_modify_block(block):
            raise Exception(
                "Cannot move blocks that are not contained in the matched operation."
            )
        if not self._can_modify_op(op):
            raise Exception(
                "Cannot move block elsewhere than before the matched operation,"
                " or before an operation child")
        Rewriter.inline_block_before(block, op)

    def inline_block_after(self, block: Block, op: Operation):
        """
        Move the block operations after the given operation.
        The block should not be a parent of the operation, and should be a child of the
        matched operation.
        The operation should also be a child of the matched operation.
        """
        self.has_done_action = True
        if op is self.current_operation:
            return self.inline_block_before_matched_op(block)
        if not self._can_modify_block(block) or (
                op.parent and not self._can_modify_block(op.parent)):
            raise Exception(
                "Cannot move blocks that are not contained in the matched operation."
            )
        Rewriter.inline_block_after(block, op)

    def move_region_contents_to_new_regions(self, region: Region) -> Region:
        """
        Move the region blocks to a new region.
        The region should be a child of the matched operation.
        """
        self.has_done_action = True
        if not self._can_modify_region(region):
            raise Exception(
                "Cannot move regions that are not children of the matched operation"
            )
        return Rewriter.move_region_contents_to_new_regions(region)


class RewritePattern(ABC):
    """
    A side-effect free rewrite pattern matching on a DAG.
    """

    @abstractmethod
    def match_and_rewrite(self, op: Operation, rewriter: PatternRewriter):
        """
        Match an operation, and optionally perform a rewrite using the rewriter.
        """
        ...


@dataclass(eq=False, repr=False)
class AnonymousRewritePattern(RewritePattern):
    """
    A rewrite pattern encoded by an anonymous function.
    """
    func: Callable[[Operation, PatternRewriter], None]

    def match_and_rewrite(self, op: Operation,
                          rewriter: PatternRewriter) -> None:
        self.func(op, rewriter)


def op_type_rewrite_pattern(
    func: Callable[..., None]
) -> Callable[[Any, Any], None] | Callable[[Any, Any, Any], None]:
    """
    This function is intended to be used as a decorator on a RewritePatter method.
    It uses type hints to match on a specific operation type before calling the decorated
    function.
    """
    # Get the operation argument and check that it is a subclass of Operation
    params = [param for param in inspect.signature(func).parameters.values()]
    if len(params) < 2:
        raise Exception(
            "op_type_rewrite_pattern expects the decorated function to "
            "have two non-self arguments.")
    is_method = params[0].name == "self"
    if is_method:
        if len(params) != 3:
            raise Exception(
                "op_type_rewrite_pattern expects the decorated method to "
                "have two non-self arguments.")
    else:
        if len(params) != 2:
            raise Exception(
                "op_type_rewrite_pattern expects the decorated function to "
                "have two arguments.")
    expected_type = params[-2].annotation
    if not issubclass(expected_type, Operation):
        raise Exception(
            "op_type_rewrite_pattern expects the first non-self argument"
            "type hint to be an Operation subclass")

    if not is_method:

        def op_type_rewrite_pattern_static_wrapper(
                op: Operation, rewriter: PatternRewriter) -> None:
            if not isinstance(op, expected_type):
                return None
            func(op, rewriter)

        return op_type_rewrite_pattern_static_wrapper

    def op_type_rewrite_pattern_method_wrapper(
            self,  # type: ignore
            op: Operation,
            rewriter: PatternRewriter) -> None:
        if not isinstance(op, expected_type):
            return None
        func(self, op, rewriter)

    return op_type_rewrite_pattern_method_wrapper  # type: ignore


@dataclass(eq=False, repr=False)
class GreedyRewritePatternApplier(RewritePattern):
    """
    Apply a list of patterns in order until one pattern matches,
    and then use this rewrite.
    """

    rewrite_patterns: list[RewritePattern]
    """The list of rewrites to apply in order."""

    def match_and_rewrite(self, op: Operation,
                          rewriter: PatternRewriter) -> None:
        for pattern in self.rewrite_patterns:
            pattern.match_and_rewrite(op, rewriter)
            if rewriter.has_done_action:
                return
        return


@dataclass(eq=False, repr=False)
class PatternRewriteWalker:
    """
    Walks the IR in the block and instruction order, and rewrite it in place.
    Previous references to the walked operations are invalid after the walk.
    Can walk either first the regions, or first the owner operation.
    The walker will also walk recursively on the created operations.
    """

    pattern: RewritePattern
    """Pattern to apply during the walk."""

    walk_regions_first: bool = field(default=False)
    """
    Choose if the walker should first walk the operation regions first,
    or the operation itself.
    """

    apply_recursively: bool = field(default=True)
    """Apply recursively rewrites on new operations."""

    walk_reverse: bool = field(default=False)
    """
    Walk the regions and blocks in reverse order.
    That way, all uses are replaced before the definitions.
    """

    def rewrite_module(self, op: ModuleOp):
        """Rewrite an entire module operation."""
        self._rewrite_op(op)

    def _rewrite_op(self, op: Operation) -> int:
        """
        Rewrite an operation, along with its regions.
        Returns by how much operations the walker should move.
        """
        # First, we rewrite the regions if needed
        if self.walk_regions_first:
            self._rewrite_op_regions(op)

        # We then match for a pattern in the current operation
        rewriter = PatternRewriter(op)
        self.pattern.match_and_rewrite(op, rewriter)

        if rewriter.has_done_action:
            # If we produce new operations, we rewrite them recursively if requested
            if self.apply_recursively:
                return (len(rewriter.added_operations_before) +
                        len(rewriter.added_operations_after) -
                        int(rewriter.has_erased_matched_operation)
                        if self.walk_reverse else 0)
            # Else, we rewrite only their regions if they are supposed to be
            # rewritten after
            else:
                if not self.walk_regions_first:
                    for new_op in rewriter.added_operations_before:
                        self._rewrite_op_regions(new_op)
                    if not rewriter.has_erased_matched_operation:
                        self._rewrite_op_regions(op)
                    for new_op in rewriter.added_operations_after:
                        self._rewrite_op_regions(new_op)
                return -1 if self.walk_reverse else len(
                    rewriter.added_operations_before) + len(
                        rewriter.added_operations_after) + int(
                            not rewriter.has_erased_matched_operation)

        # Otherwise, we only rewrite the regions of the operation if needed
        if not self.walk_regions_first:
            self._rewrite_op_regions(op)
        return -1 if self.walk_reverse else 1

    def _rewrite_op_regions(self, op: Operation):
        """
        Rewrite the regions of an operation, and update the operation with the
        new regions.
        """
        if not self.walk_reverse:
            for region in op.regions:
                for block in region.blocks:
                    idx = 0
                    while idx < len(block.ops):
                        idx += self._rewrite_op(block.ops[idx])
        else:
            for region in op.regions:
                for block in reversed(region.blocks):
                    idx = len(block.ops) - 1
                    while idx >= 0:
                        idx += self._rewrite_op(block.ops[idx])



import versioneer
from setuptools import find_packages, setup
from pathlib import Path

# Add README.md as long description
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

with open("requirements.txt") as f:
    required = f.read().splitlines()

with open("requirements-optional.txt") as f:
    optionals = f.read().splitlines()

reqs = []
for ir in required:
    if ir[0:3] == "git":
        name = ir.split("/")[-1]
        reqs += ["%s @ %s@main" % (name, ir)]
    else:
        reqs += [ir]

extras_require = {}
for mreqs, mode in zip(
    [
        optionals,
    ],
    [
        "extras",
    ],
):
    opt_reqs = []
    for ir in mreqs:
        # For conditionals like pytest=2.1; python == 3.6
        if ";" in ir:
            entries = ir.split(";")
            extras_require[entries[1]] = entries[0]
        # Git repos, install master
        if ir[0:3] == "git":
            name = ir.split("/")[-1]
            opt_reqs += ["%s @ %s@main" % (name, ir)]
        else:
            opt_reqs += [ir]
    extras_require[mode] = opt_reqs

setup(
    name="xdsl",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="xDSL",
    long_description=long_description,
    long_description_content_type='text/markdown',
    project_urls={
        "Source Code": "https://github.com/xdslproject/xdsl",
        "Issue Tracker": "https://github.com/xdslproject/xdsl/issues",
    },
    url="https://xdsl.dev/",
    platforms=["Linux", "Mac OS-X", "Unix"],
    test_suite="pytest",
    author="Mathieu Fehr",
    author_email="mathieu.fehr@ed.ac.uk",
    license="MIT",
    packages=find_packages(),
    package_data={"xdsl": ["py.typed"]},
    install_requires=reqs,
    extras_require=extras_require,
    zip_safe=False,
)



from __future__ import annotations
from typing import Annotated, Union

from xdsl.dialects.builtin import Float32Type, Float64Type
from xdsl.ir import (MLIRType, ParametrizedAttribute, Operation, Dialect,
                     OpResult, SSAValue)
from xdsl.irdl import (irdl_op_definition, irdl_attr_definition, Operand,
                       ParameterDef, ParamAttrConstraint, AnyOf,
                       VerifyException)


@irdl_attr_definition
class ComplexType(ParametrizedAttribute, MLIRType):
    name = "cmath.complex"
    data: ParameterDef[Float64Type | Float32Type]


@irdl_op_definition
class Norm(Operation):
    name: str = "cmath.norm"

    op: Annotated[
        Operand,
        ParamAttrConstraint(ComplexType, [AnyOf([Float32Type, Float64Type])])]
    res: Annotated[OpResult, AnyOf([Float32Type, Float64Type])]

    # TODO replace with trait
    def verify_(self) -> None:
        if self.op.typ.data != self.res.typ:
            raise VerifyException(
                "expect all input and output types to be equal")


@irdl_op_definition
class Mul(Operation):
    name: str = "cmath.mul"

    lhs: Annotated[
        Operand,
        ParamAttrConstraint(ComplexType, [AnyOf([Float32Type, Float64Type])])]
    rhs: Annotated[
        Operand,
        ParamAttrConstraint(ComplexType, [AnyOf([Float32Type, Float64Type])])]
    result: Annotated[
        OpResult,
        ParamAttrConstraint(ComplexType, [AnyOf([Float32Type, Float64Type])])]

    # TODO replace with trait
    def verify_(self) -> None:
        if self.lhs != self.rhs.typ and self.rhs.typ != self.result.typ:
            raise VerifyException(
                "expect all input and output types to be equal")

    @staticmethod
    def get(operand1: Union[Operation, SSAValue],
            operand2: Union[Operation, SSAValue]) -> Mul:
        operand1 = SSAValue.get(operand1)
        return Mul.build(operands=[operand1, operand2],
                         result_types=[operand1.typ])


CMath = Dialect([Norm, Mul], [ComplexType])



from xdsl.utils.hints import is_satisfying_hint
from typing import Any, TypeAlias


class Class1:
    pass


class SubClass1(Class1):
    pass


class Class2:
    pass


#     _
#    / \   _ __  _   _
#   / _ \ | '_ \| | | |
#  / ___ \| | | | |_| |
# /_/   \_\_| |_|\__, |
#                |___/
#


def test_any_hint():
    """Test that the we can check if a value is satisfying `Any`."""
    assert is_satisfying_hint(3, Any)
    assert is_satisfying_hint([], Any)
    assert is_satisfying_hint([2], Any)
    assert is_satisfying_hint(int, Any)


#   ____ _
#  / ___| | __ _ ___ ___
# | |   | |/ _` / __/ __|
# | |___| | (_| \__ \__ \
#  \____|_|\__,_|___/___/
#


def test_class_hint_correct():
    """Test that a class hint is being satisfied by giving a class instance."""
    assert is_satisfying_hint(Class1(), Class1)
    assert is_satisfying_hint(SubClass1(), SubClass1)
    assert is_satisfying_hint(Class2(), Class2)
    assert is_satisfying_hint(True, bool)
    assert is_satisfying_hint(10, int)
    assert is_satisfying_hint("", str)


def test_class_hint_class():
    """Test that a class hint is not being satisfied by the class object."""
    assert not is_satisfying_hint(Class1, Class1)


def test_class_hint_wrong_instance():
    """
    Test that a class hint is not being satisfied by another class instance.
    """
    assert not is_satisfying_hint(Class2(), Class1)
    assert not is_satisfying_hint(0, Class1)
    assert not is_satisfying_hint("", Class1)
    assert not is_satisfying_hint("", int)


def test_class_hint_subclass():
    """
    Test that a class hint is satisfied by a subclass.
    """
    assert is_satisfying_hint(SubClass1(), Class1)
    assert is_satisfying_hint(True, int)  # bool is a subclass of int


#  _     _     _
# | |   (_)___| |_
# | |   | / __| __|
# | |___| \__ \ |_
# |_____|_|___/\__|
#


def test_list_hint_empty():
    """Test that empty lists satisfy all list hints."""
    assert is_satisfying_hint([], list[int])
    assert is_satisfying_hint([], list[bool])
    assert is_satisfying_hint([], list[Class1])


def test_list_hint_correct():
    """
    Test that list hints work correcly on non-empty lists of the right type.
    """
    assert is_satisfying_hint([42], list[int])
    assert is_satisfying_hint([0, 3, 5], list[int])
    assert is_satisfying_hint([False], list[bool])
    assert is_satisfying_hint([True, False], list[bool])
    assert is_satisfying_hint([Class1(), SubClass1()], list[Class1])


def test_list_hint_not_list_failure():
    """Test that list hints work correcly on non lists."""
    assert not is_satisfying_hint(0, list[int])
    assert not is_satisfying_hint(0, list[Any])
    assert not is_satisfying_hint(True, list[bool])
    assert not is_satisfying_hint(True, list[Any])
    assert not is_satisfying_hint("", list[Any])
    assert not is_satisfying_hint("", list[str])
    assert not is_satisfying_hint(Class1(), list[Class1])
    assert not is_satisfying_hint(Class1(), list[Any])
    assert not is_satisfying_hint({}, list[dict[Any, Any]])
    assert not is_satisfying_hint({}, list[Any])


def test_list_hint_failure():
    """
    Test that list hints work correcly on non-empty lists of the wrong type.
    """
    assert not is_satisfying_hint([0], list[bool])
    assert not is_satisfying_hint([0, True], list[bool])
    assert not is_satisfying_hint([True, 0], list[bool])
    assert not is_satisfying_hint([True, False, True, 0], list[bool])
    assert not is_satisfying_hint([Class2()], list[Class1])


def test_list_hint_nested():
    """
    Test that we can check nested list hints.
    """
    assert is_satisfying_hint([[]], list[list[int]])
    assert is_satisfying_hint([[0]], list[list[int]])
    assert is_satisfying_hint([[0], [2, 1]], list[list[int]])
    assert not is_satisfying_hint([[0], [2, 1], [""]], list[list[int]])
    assert not is_satisfying_hint([[0], [2, 1], [32, ""]], list[list[int]])
    assert is_satisfying_hint([], list[list[list[int]]])
    assert is_satisfying_hint([[]], list[list[list[int]]])
    assert is_satisfying_hint([[[]]], list[list[list[int]]])
    assert is_satisfying_hint([[], [[]]], list[list[list[int]]])
    assert is_satisfying_hint([[], [[0, 32]]], list[list[list[int]]])


#   ____  _      _
# |  _ \(_) ___| |_
# | | | | |/ __| __|
# | |_| | | (__| |_
# |____/|_|\___|\__|
#


def test_dict_hint_empty():
    """Test that empty dicts satisfy all dict hints."""
    assert is_satisfying_hint({}, dict[int, bool])
    assert is_satisfying_hint({}, dict[Any, Any])
    assert is_satisfying_hint({}, dict[Class1, Any])
    assert is_satisfying_hint({}, dict[str, int])


def test_dict_hint_correct():
    """
    Test that dict hints work correcly on non-empty dicts of the right type.
    """
    assert is_satisfying_hint({"": 0}, dict[str, int])
    assert is_satisfying_hint({"": 0, "a": 32}, dict[str, int])
    assert is_satisfying_hint({0: "", 32: "a"}, dict[int, str])
    assert is_satisfying_hint({
        True: Class1(),
        False: SubClass1()
    }, dict[bool, Class1])


def test_dict_hint_not_dict_failure():
    """Test that dict hints work correcly on non dicts."""
    assert not is_satisfying_hint(0, dict[int, int])
    assert not is_satisfying_hint(0, dict[Any, Any])
    assert not is_satisfying_hint(True, dict[bool, bool])
    assert not is_satisfying_hint(True, dict[Any, Any])
    assert not is_satisfying_hint("", dict[Any, Any])
    assert not is_satisfying_hint(Class1(), dict[Any, Any])
    assert not is_satisfying_hint([], dict[Any, Any])


def test_dict_hint_failure():
    """
    Test that dict hints work correcly on non-empty dicts of the wrong type.
    """
    assert not is_satisfying_hint({"": ""}, dict[int, str])
    assert not is_satisfying_hint({0: 0}, dict[int, str])
    assert not is_satisfying_hint({0: "0", 2: 2}, dict[int, str])
    assert not is_satisfying_hint({0: "0", 2: "1", "2": "2"}, dict[int, str])


threeDict: TypeAlias = dict[int, dict[int, dict[int, str]]]


def test_dict_hint_nested():
    """
    Test that we can check nested dict hints.
    """
    assert is_satisfying_hint({}, dict[int, dict[int, str]])
    assert is_satisfying_hint({0: {}}, dict[int, dict[int, str]])
    assert is_satisfying_hint({0: {1: ""}}, dict[int, dict[int, str]])

    assert not is_satisfying_hint({"": {}}, dict[int, dict[int, str]])
    assert not is_satisfying_hint({0: ""}, dict[int, dict[int, str]])
    assert not is_satisfying_hint({0: {"": ""}}, dict[int, dict[int, str]])
    assert not is_satisfying_hint({0: {0: 0}}, dict[int, dict[int, str]])

    assert is_satisfying_hint({}, threeDict)
    assert is_satisfying_hint({0: {}}, threeDict)
    assert is_satisfying_hint({0: {0: {}}}, threeDict)
    assert is_satisfying_hint({0: {}, 1: {0: {}}}, threeDict)
    assert is_satisfying_hint({0: {}, 1: {0: {0: "0", 1: "32"}}}, threeDict)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import (TypeAlias, List, cast, Type, Sequence, TYPE_CHECKING, Any,
                    TypeVar)

from xdsl.ir import (Data, MLIRType, ParametrizedAttribute, Operation, Region,
                     Attribute, Dialect)
from xdsl.irdl import (OpAttr, VarOpResult, VarOperand, VarRegion,
                       irdl_attr_definition, attr_constr_coercion,
                       irdl_data_definition, irdl_to_attr_constraint,
                       irdl_op_definition, builder, ParameterDef,
                       SingleBlockRegion, Generic, GenericData, AttrConstraint,
                       AnyAttr)
from xdsl.utils.exceptions import VerifyException

if TYPE_CHECKING:
    from xdsl.parser import BaseParser
    from utils.exceptions import ParseError
    from xdsl.printer import Printer


@irdl_attr_definition
class StringAttr(Data[str]):
    name = "string"

    @staticmethod
    def parse_parameter(parser: BaseParser) -> str:
        data = parser.parse_str_literal()
        return data

    @staticmethod
    def print_parameter(data: str, printer: Printer) -> None:
        printer.print_string(f'"{data}"')

    @staticmethod
    @builder
    def from_str(data: str) -> StringAttr:
        return StringAttr(data)

    @staticmethod
    @builder
    def from_int(data: int) -> StringAttr:
        return StringAttr(str(data))


@irdl_attr_definition
class SymbolNameAttr(ParametrizedAttribute):
    name = "symbol_name"
    data: ParameterDef[StringAttr]

    @staticmethod
    @builder
    def from_str(data: str) -> SymbolNameAttr:
        return SymbolNameAttr([StringAttr.from_str(data)])

    @staticmethod
    @builder
    def from_string_attr(data: StringAttr) -> SymbolNameAttr:
        return SymbolNameAttr([data])


@irdl_attr_definition
class FlatSymbolRefAttr(ParametrizedAttribute):
    name = "flat_symbol_ref"
    data: ParameterDef[StringAttr]

    @staticmethod
    @builder
    def from_str(data: str) -> FlatSymbolRefAttr:
        return FlatSymbolRefAttr([StringAttr(data)])

    @staticmethod
    @builder
    def from_string_attr(data: StringAttr) -> FlatSymbolRefAttr:
        return FlatSymbolRefAttr([data])


@irdl_attr_definition
class IntAttr(Data[int]):
    name = "int"

    @staticmethod
    def parse_parameter(parser: BaseParser) -> int:
        data = parser.parse_int_literal()
        return data

    @staticmethod
    def print_parameter(data: int, printer: Printer) -> None:
        printer.print_string(f'{data}')

    @staticmethod
    @builder
    def from_int(data: int) -> IntAttr:
        return IntAttr(data)


class Signedness(Enum):
    "Signedness semantics for integer"

    SIGNLESS = 0
    "No signedness semantics"

    SIGNED = 1
    UNSIGNED = 2


@irdl_data_definition
class SignednessAttr(Data[Signedness]):
    name = "signedness"

    @staticmethod
    def parse_parameter(parser: BaseParser) -> Signedness:
        if parser.parse_optional_string("signless") is not None:
            return Signedness.SIGNLESS
        elif parser.parse_optional_string("signed") is not None:
            return Signedness.SIGNED
        elif parser.parse_optional_string("unsigned") is not None:
            return Signedness.UNSIGNED
        raise ParseError(parser.get_pos(), "Expected signedness")

    @staticmethod
    def print_parameter(data: Signedness, printer: Printer) -> None:
        if data == Signedness.SIGNLESS:
            printer.print_string("signless")
        elif data == Signedness.SIGNED:
            printer.print_string("signed")
        elif data == Signedness.UNSIGNED:
            printer.print_string("unsigned")
        else:
            raise ValueError(f"Invalid signedness {data}")

    @staticmethod
    @builder
    def from_enum(signedness: Signedness) -> SignednessAttr:
        return SignednessAttr(signedness)


@irdl_attr_definition
class IntegerType(ParametrizedAttribute):
    name = "integer_type"
    width: ParameterDef[IntAttr]
    signedness: ParameterDef[SignednessAttr]

    @staticmethod
    @builder
    def from_width(
            width: int,
            signedness: Signedness = Signedness.SIGNLESS) -> IntegerType:
        return IntegerType(
            [IntAttr.from_int(width),
             SignednessAttr.from_enum(signedness)])


i64 = IntegerType.from_width(64)
i32 = IntegerType.from_width(32)
i1 = IntegerType.from_width(1)


@irdl_attr_definition
class UnitAttr(ParametrizedAttribute):
    name = "unit"


@irdl_attr_definition
class IndexType(ParametrizedAttribute):
    name = "index"


_IntegerAttrTyp = TypeVar("_IntegerAttrTyp",
                          bound=IntegerType | IndexType,
                          covariant=True)


@irdl_attr_definition
class IntegerAttr(Generic[_IntegerAttrTyp], ParametrizedAttribute):
    name = "integer"
    value: ParameterDef[IntAttr]
    typ: ParameterDef[_IntegerAttrTyp]

    @staticmethod
    @builder
    def from_int_and_width(value: int, width: int) -> IntegerAttr[IntegerType]:
        return IntegerAttr(
            [IntAttr.from_int(value),
             IntegerType.from_width(width)])

    @staticmethod
    @builder
    def from_index_int_value(value: int) -> IntegerAttr[IndexType]:
        return IntegerAttr([IntAttr.from_int(value), IndexType()])

    @staticmethod
    @builder
    def from_params(
            value: int | IntAttr,
            typ: int | Attribute) -> IntegerAttr[IntegerType | IndexType]:
        value = IntAttr.build(value)
        if not isinstance(typ, IndexType):
            typ = IntegerType.build(typ)
        return IntegerAttr([value, typ])


AnyIntegerAttr: TypeAlias = IntegerAttr[IntegerType | IndexType]


@irdl_attr_definition
class Float16Type(ParametrizedAttribute, MLIRType):
    name = "f16"


@irdl_attr_definition
class Float32Type(ParametrizedAttribute, MLIRType):
    name = "f32"


class Float64Type(ParametrizedAttribute, MLIRType):
    name = "f64"


AnyFloat: TypeAlias = Float16Type | Float32Type | Float64Type


@irdl_attr_definition
class FloatData(Data[float]):
    name = "float_data"

    @staticmethod
    def parse_parameter(parser: BaseParser) -> float:
        return parser.parse_float_literal()

    @staticmethod
    def print_parameter(data: float, printer: Printer) -> None:
        printer.print_string(f'{data}')

    @staticmethod
    @builder
    def from_float(data: float) -> FloatData:
        return FloatData(data)


_FloatAttrTyp = TypeVar("_FloatAttrTyp", bound=AnyFloat, covariant=True)

_FloatAttrTypContr = TypeVar("_FloatAttrTypContr", bound=AnyFloat)


@irdl_attr_definition
class FloatAttr(Generic[_FloatAttrTyp], ParametrizedAttribute):
    name = "float"

    value: ParameterDef[FloatData]
    type: ParameterDef[_FloatAttrTyp]

    @staticmethod
    @builder
    def from_value(
        value: float, type: _FloatAttrTypContr = Float32Type()
    ) -> FloatAttr[_FloatAttrTypContr]:
        return FloatAttr([FloatData.from_float(value), type])

    @staticmethod
    @builder
    def from_float_and_width(value: float, width: int) -> FloatAttr[AnyFloat]:
        if width == 16:
            return FloatAttr([FloatData.from_float(value), Float16Type()])
        if width == 32:
            return FloatAttr([FloatData.from_float(value), Float32Type()])
        if width == 64:
            return FloatAttr([FloatData.from_float(value), Float64Type()])
        raise ValueError(f"Invalid bitwidth: {width}")


AnyFloatAttr: TypeAlias = FloatAttr[AnyFloat]


@dataclass
class ArrayOfConstraint(AttrConstraint):
    """
    A constraint that enforces an ArrayData whose elements all satisfy
    the elem_constr.
    """
    elem_constr: AttrConstraint

    def __init__(self, constr: Attribute | Type[Attribute] | AttrConstraint):
        self.elem_constr = attr_constr_coercion(constr)

    def verify(self, attr: Attribute) -> None:
        if not isinstance(attr, Data):
            raise Exception(f"expected data ArrayData but got {attr}")
        for e in cast(ArrayAttr[Attribute], attr).data:
            self.elem_constr.verify(e)


_ArrayAttrT = TypeVar("_ArrayAttrT", bound=Attribute, covariant=True)


@irdl_attr_definition
class ArrayAttr(GenericData[List[_ArrayAttrT]]):
    name = "array"

    @staticmethod
    def parse_parameter(parser: BaseParser) -> List[_ArrayAttrT]:
        parser.parse_char("[")
        data = parser.parse_list(parser.parse_optional_attribute)
        parser.parse_char("]")
        # the type system can't ensure that the elements are of type A
        # and not just of type Attribute, therefore, the following cast
        return cast(List[_ArrayAttrT], data)

    @staticmethod
    def print_parameter(data: List[_ArrayAttrT], printer: Printer) -> None:
        printer.print_string("[")
        printer.print_list(data, printer.print_attribute)
        printer.print_string("]")

    @staticmethod
    def generic_constraint_coercion(args: tuple[Any]) -> AttrConstraint:
        if len(args) == 1:
            return ArrayOfConstraint(irdl_to_attr_constraint(args[0]))
        if len(args) == 0:
            return ArrayOfConstraint(AnyAttr())
        raise TypeError(f"Attribute ArrayAttr expects at most type"
                        f" parameter, but {len(args)} were given")

    def verify(self) -> None:
        if not isinstance(self.data, list):
            raise VerifyException(
                f"Wrong type given to attribute {self.name}: got"
                f" {type(self.data)}, but expected list of"
                " attributes")
        for idx, val in enumerate(self.data):
            if not isinstance(val, Attribute):
                raise VerifyException(
                    f"{self.name} data expects attribute list, but {idx} "
                    f"element is of type {type(val)}")

    @staticmethod
    @builder
    def from_list(data: List[_ArrayAttrT]) -> ArrayAttr[_ArrayAttrT]:
        return ArrayAttr(data)


AnyArrayAttr: TypeAlias = ArrayAttr[Attribute]


@irdl_attr_definition
class DictionaryAttr(GenericData[dict[str, Attribute]]):
    name = "dictionary"

    @staticmethod
    def parse_parameter(parser: BaseParser) -> dict[str, Attribute]:
        # force MLIR style parsing of attribute
        from xdsl.parser import MLIRParser
        return MLIRParser.parse_optional_attr_dict(parser)

    @staticmethod
    def print_parameter(data: dict[str, Attribute], printer: Printer) -> None:
        printer.print_string("{")
        printer.print_dictionary(data, printer.print_string_literal,
                                 printer.print_attribute)
        printer.print_string("}")

    @staticmethod
    def generic_constraint_coercion(args: tuple[Any]) -> AttrConstraint:
        raise Exception(f"Unsupported operation on {DictionaryAttr.name}")

    def verify(self) -> None:
        if not isinstance(self.data, dict):
            raise VerifyException(
                f"Wrong type given to attribute {self.name}: got"
                f" {type(self.data)}, but expected dictionary of"
                " attributes")
        for key, val in self.data.items():
            if not isinstance(key, str):
                raise VerifyException(
                    f"{self.name} key expects str, but {key} "
                    f"element is of type {type(key)}")
            if not isinstance(val, Attribute):
                raise VerifyException(
                    f"{self.name} key expects attribute, but {val} "
                    f"element is of type {type(val)}")

    @staticmethod
    @builder
    def from_dict(data: dict[str | StringAttr, Attribute]) -> DictionaryAttr:
        to_add_data: dict[str, Attribute] = {}
        for k, v in data.items():
            # try to coerce keys into StringAttr
            if isinstance(k, StringAttr):
                k = k.data
            # if coercion fails, raise KeyError!
            if not isinstance(k, str):
                raise TypeError(
                    f"DictionaryAttr.from_dict expects keys to"
                    f" be of type str or StringAttr, but {type(k)} provided")
            to_add_data[k] = v
        return DictionaryAttr(to_add_data)


@irdl_attr_definition
class TupleType(ParametrizedAttribute):
    name = "tuple"

    types: ParameterDef[ArrayAttr[Attribute]]

    @staticmethod
    @builder
    def from_type_list(types: List[Attribute]) -> TupleType:
        return TupleType([ArrayAttr.from_list(types)])


_VectorTypeElems = TypeVar("_VectorTypeElems", bound=Attribute)


@irdl_attr_definition
class VectorType(Generic[_VectorTypeElems], ParametrizedAttribute, MLIRType):
    name = "vector"

    shape: ParameterDef[ArrayAttr[AnyIntegerAttr]]
    element_type: ParameterDef[_VectorTypeElems]

    def get_num_dims(self) -> int:
        return len(self.shape.data)

    def get_shape(self) -> List[int]:
        return [i.value.data for i in self.shape.data]

    @staticmethod
    @builder
    def from_element_type_and_shape(
        referenced_type: _VectorTypeElems,
        shape: List[int | IntegerAttr[IndexType]]
    ) -> VectorType[_VectorTypeElems]:
        return VectorType([
            ArrayAttr.from_list(
                [IntegerAttr[IntegerType].build(d) for d in shape]),
            referenced_type
        ])

    @staticmethod
    @builder
    def from_params(
        referenced_type: _VectorTypeElems,
        shape: ArrayAttr[IntegerAttr[IntegerType]] = ArrayAttr.from_list(
            [IntegerAttr.from_int_and_width(1, 64)])
    ) -> VectorType[_VectorTypeElems]:
        return VectorType([shape, referenced_type])


AnyVectorType: TypeAlias = VectorType[Attribute]

_TensorTypeElems = TypeVar("_TensorTypeElems", bound=Attribute, covariant=True)


@irdl_attr_definition
class TensorType(Generic[_TensorTypeElems], ParametrizedAttribute, MLIRType):
    name = "tensor"

    shape: ParameterDef[ArrayAttr[AnyIntegerAttr]]
    element_type: ParameterDef[_TensorTypeElems]

    def get_num_dims(self) -> int:
        return len(self.shape.data)

    def get_shape(self) -> List[int]:
        return [i.value.data for i in self.shape.data]

    @staticmethod
    @builder
    def from_type_and_list(
        referenced_type: _TensorTypeElems,
        shape: Sequence[int | IntegerAttr[IndexType]] | None = None
    ) -> TensorType[_TensorTypeElems]:
        if shape is None:
            shape = [1]
        return TensorType([
            ArrayAttr.from_list(
                [IntegerAttr[IndexType].build(d) for d in shape]),
            referenced_type
        ])

    @staticmethod
    @builder
    def from_params(
        referenced_type: _VectorTypeElems,
        shape: AnyArrayAttr = AnyArrayAttr.from_list(
            [IntegerAttr.from_int_and_width(1, 64)])
    ) -> TensorType[_VectorTypeElems]:
        return TensorType([shape, referenced_type])


AnyTensorType: TypeAlias = TensorType[Attribute]

_UnrankedTensorTypeElems = TypeVar("_UnrankedTensorTypeElems",
                                   bound=Attribute,
                                   covariant=True)


@irdl_attr_definition
class UnrankedTensorType(Generic[_UnrankedTensorTypeElems],
                         ParametrizedAttribute, MLIRType):
    name = "unranked_tensor"

    element_type: ParameterDef[_UnrankedTensorTypeElems]

    @staticmethod
    @builder
    def from_type(
        referenced_type: _UnrankedTensorTypeElems
    ) -> UnrankedTensorType[_UnrankedTensorTypeElems]:
        return UnrankedTensorType([referenced_type])


AnyUnrankedTensorType: TypeAlias = UnrankedTensorType[Attribute]


@dataclass(init=False)
class ContainerOf(AttrConstraint):
    """A type constraint that can be nested once in a vector or a tensor."""
    elem_constr: AttrConstraint

    def __init__(
            self,
            elem_constr: Attribute | type[Attribute] | AttrConstraint) -> None:
        self.elem_constr = attr_constr_coercion(elem_constr)

    def verify(self, attr: Attribute) -> None:
        if isinstance(attr, VectorType) or isinstance(attr, TensorType):
            self.elem_constr.verify(attr.element_type)  # type: ignore
        else:
            self.elem_constr.verify(attr)


_VectorOrTensorElem = TypeVar("_VectorOrTensorElem", bound=Attribute)

VectorOrTensorOf: TypeAlias = (VectorType[_VectorOrTensorElem]
                               | TensorType[_VectorOrTensorElem]
                               | UnrankedTensorType[_VectorOrTensorElem])


@irdl_attr_definition
class DenseIntOrFPElementsAttr(ParametrizedAttribute):
    name = "dense"
    type: ParameterDef[VectorOrTensorOf[IntegerType]
                       | VectorOrTensorOf[IndexType]
                       | VectorOrTensorOf[AnyFloat]]
    data: ParameterDef[ArrayAttr[AnyIntegerAttr] | ArrayAttr[AnyFloatAttr]]

    # The type stores the shape data
    @property
    def shape(self) -> List[int]:
        return self.type.get_shape()

    @property
    def shape_is_complete(self) -> bool:
        shape = self.shape
        if not len(shape):
            return False

        n = 1
        for dim in shape:
            if dim < 1:
                # Dimensions need to be greater or equal to one
                return False
            n *= dim

        # Product of dimensions needs to equal length
        return n == len(self.data.data)

    @staticmethod
    @builder
    def create_dense_index(
            type: VectorOrTensorOf[IndexType],
            data: List[int | IntegerAttr[IndexType]]
    ) -> DenseIntOrFPElementsAttr:
        attr_list = [
            IntegerAttr.from_index_int_value(d) if isinstance(d, int) else d
            for d in data
        ]
        return DenseIntOrFPElementsAttr([type, ArrayAttr.from_list(attr_list)])

    @staticmethod
    @builder
    def create_dense_int(
        type: VectorOrTensorOf[IntegerType],
        data: List[int | IntegerAttr[IntegerType]]
    ) -> DenseIntOrFPElementsAttr:
        attr_list = [
            IntegerAttr.from_params(d, type.element_type) if isinstance(
                d, int) else d for d in data
        ]
        return DenseIntOrFPElementsAttr([type, ArrayAttr.from_list(attr_list)])

    @staticmethod
    @builder
    def create_dense_float(
            type: VectorOrTensorOf[AnyFloat],
            data: List[int | float | AnyFloatAttr]
    ) -> DenseIntOrFPElementsAttr:
        data_attr = [
            FloatAttr.from_value(float(d), type.element_type)
            if not isinstance(d, FloatAttr) else d for d in data
        ]
        return DenseIntOrFPElementsAttr([type, ArrayAttr.from_list(data_attr)])

    @staticmethod
    @builder
    def from_list(
        type: VectorOrTensorOf[Attribute], data: List[int | AnyIntegerAttr]
        | List[int | float | AnyFloatAttr]
    ) -> DenseIntOrFPElementsAttr:
        if isinstance(type.element_type, IntegerType):
            return DenseIntOrFPElementsAttr.create_dense_int(type, data)
        elif isinstance(type.element_type, IndexType):
            return DenseIntOrFPElementsAttr.create_dense_index(type, data)
        elif isinstance(type.element_type, AnyFloat):
            return DenseIntOrFPElementsAttr.create_dense_float(type, data)
        else:
            raise TypeError(f"Unsupported element type {type.element_type}")

    @staticmethod
    @builder
    def vector_from_list(
            data: List[int] | List[float],
            typ: IntegerType | IndexType | AnyFloat
    ) -> DenseIntOrFPElementsAttr:
        t = AnyVectorType.from_element_type_and_shape(typ, [len(data)])
        return DenseIntOrFPElementsAttr.from_list(t, data)

    @staticmethod
    @builder
    def tensor_from_list(data: List[int] | List[float],
                         typ: IntegerType | IndexType | AnyFloat,
                         shape: List[int] = []) -> DenseIntOrFPElementsAttr:
        t = AnyTensorType.from_type_and_list(
            typ, shape if len(shape) else [len(data)])
        return DenseIntOrFPElementsAttr.from_list(t, data)


@irdl_attr_definition
class FunctionType(ParametrizedAttribute, MLIRType):
    name = "fun"

    inputs: ParameterDef[ArrayAttr[Attribute]]
    outputs: ParameterDef[ArrayAttr[Attribute]]

    @staticmethod
    @builder
    def from_lists(inputs: List[Attribute],
                   outputs: List[Attribute]) -> FunctionType:
        return FunctionType(
            [ArrayAttr.from_list(inputs),
             ArrayAttr.from_list(outputs)])

    @staticmethod
    @builder
    def from_attrs(inputs: ArrayAttr[Attribute],
                   outputs: ArrayAttr[Attribute]) -> FunctionType:
        return FunctionType([inputs, outputs])


@irdl_attr_definition
class NoneAttr(ParametrizedAttribute):
    """An attribute representing the absence of an attribute."""
    name: str = "none"


@irdl_attr_definition
class OpaqueAttr(ParametrizedAttribute):
    name: str = "opaque"

    ident: ParameterDef[StringAttr]
    value: ParameterDef[StringAttr]
    type: ParameterDef[Attribute]

    @staticmethod
    def from_strings(name: str, value: str,
                     type: Attribute = NoneAttr()) -> OpaqueAttr:
        return OpaqueAttr(
            [StringAttr.from_str(name),
             StringAttr.from_str(value), type])


@irdl_op_definition
class UnregisteredOp(Operation):
    name: str = "builtin.unregistered"

    op_name__: OpAttr[StringAttr]
    args: VarOperand
    res: VarOpResult
    regs: VarRegion

    @property
    def op_name(self) -> StringAttr:
        return self.op_name__  # type: ignore

    @classmethod
    def with_name(cls, name: str, ctx: MLContext) -> type[Operation]:
        if name in ctx.registered_unregistered_ops:
            return ctx.registered_unregistered_ops[name]  # type: ignore

        class UnregisteredOpWithName(UnregisteredOp):

            @classmethod
            def create(cls, **kwargs):
                op = super().create(**kwargs)
                op.attributes['op_name__'] = StringAttr.build(name)
                return op

        ctx.registered_unregistered_ops[name] = UnregisteredOpWithName
        return UnregisteredOpWithName


@irdl_op_definition
class ModuleOp(Operation):
    name: str = "builtin.module"

    body: SingleBlockRegion

    @property
    def ops(self) -> List[Operation]:
        return self.regions[0].blocks[0].ops

    @staticmethod
    def from_region_or_ops(ops: List[Operation] | Region) -> ModuleOp:
        if isinstance(ops, list):
            region = Region.from_operation_list(ops)
        elif isinstance(ops, Region):
            region = ops
        else:
            raise TypeError(
                f"Expected region or operation list in ModuleOp.get, but got '{ops}'"
            )
        op = ModuleOp.create([], [], regions=[region])
        return op


# FloatXXType shortcuts
f16 = Float16Type()
f32 = Float32Type()
f64 = Float64Type()

Builtin = Dialect(
    [ModuleOp, UnregisteredOp],
    [
        StringAttr,
        FlatSymbolRefAttr,
        SymbolNameAttr,
        IntAttr,
        IntegerAttr,
        ArrayAttr,
        DictionaryAttr,
        DenseIntOrFPElementsAttr,
        UnitAttr,
        FloatData,
        NoneAttr,
        OpaqueAttr,

        # Types
        FunctionType,
        Float16Type,
        Float32Type,
        Float64Type,
        FloatAttr,
        TupleType,
        IntegerType,
        IndexType,
        VectorType,
        TensorType,
        UnrankedTensorType
    ])



from io import StringIO
from typing import Callable

import pytest

from xdsl.dialects.arith import Arith, Constant, Addi
from xdsl.dialects.builtin import ModuleOp, Builtin, i32
from xdsl.dialects.scf import Scf, Yield
from xdsl.dialects.func import Func
from xdsl.ir import MLContext, Block, SSAValue, OpResult, BlockArgument
from xdsl.parser import Parser
from xdsl.printer import Printer
from xdsl.rewriter import Rewriter


@pytest.mark.parametrize("name,result", [
    ('a', 'a'),
    ('test', 'test'),
    ('test1', None),
    ('1', None),
])
def test_ssa_value_name_hints(name, result):
    """
    The rewriter assumes, that ssa value name hints (their .name field) does not end in a numeric value. If it does,
    it will generate broken rewrites that potentially assign twice to an SSA value.

    Therefore, the SSAValue class prevents the setting of names ending in a number.
    """
    val = BlockArgument(i32, Block(), 0)

    val.name = name
    assert val.name == result



"""Unit tests for IRDL."""

from dataclasses import dataclass
import pytest

from xdsl.ir import Attribute, Data, ParametrizedAttribute
from xdsl.irdl import (AllOf, AnyAttr, AnyOf, AttrConstraint, BaseAttr,
                       EqAttrConstraint, ParamAttrConstraint, ParameterDef,
                       irdl_attr_definition)
from xdsl.parser import BaseParser
from xdsl.printer import Printer
from xdsl.utils.exceptions import VerifyException


@irdl_attr_definition
class BoolData(Data[bool]):
    """An attribute holding a boolean value."""
    name = "bool"

    @staticmethod
    def parse_parameter(parser: BaseParser) -> bool:
        raise NotImplementedError()

    @staticmethod
    def print_parameter(data: bool, printer: Printer):
        printer.print_string(str(data))


@irdl_attr_definition
class IntData(Data[int]):
    """An attribute holding an integer value."""
    name = "int"

    @staticmethod
    def parse_parameter(parser: BaseParser) -> int:
        return parser.parse_int_literal()

    @staticmethod
    def print_parameter(data: int, printer: Printer):
        printer.print_string(str(data))


@irdl_attr_definition
class DoubleParamAttr(ParametrizedAttribute):
    """An attribute with two unbounded attribute parameters."""
    name = "param"

    param1: ParameterDef[Attribute]
    param2: ParameterDef[Attribute]


def test_eq_attr_verify():
    """Check that an EqAttrConstraint verifies the expected attribute"""
    bool_true = BoolData(True)
    eq_true_constraint = EqAttrConstraint(bool_true)
    eq_true_constraint.verify(bool_true)


def test_eq_attr_verify_wrong_parameters_fail():
    """
    Check that an EqAttrConstraint fails to verify an attribute with different
    parameters.
    """
    bool_true = BoolData(True)
    bool_false = BoolData(False)
    eq_true_constraint = EqAttrConstraint(bool_true)
    with pytest.raises(VerifyException) as e:
        eq_true_constraint.verify(bool_false)
    assert e.value.args[0] == (
        f"Expected attribute {bool_true} but got {bool_false}")


def test_eq_attr_verify_wrong_base_fail():
    """
    Check that an EqAttrConstraint fails to verify an attribute with a
    different base attribute.
    """
    bool_true = BoolData(True)
    int_zero = IntData(0)
    eq_true_constraint = EqAttrConstraint(bool_true)
    with pytest.raises(VerifyException) as e:
        eq_true_constraint.verify(int_zero)
    assert e.value.args[0] == (
        f"Expected attribute {bool_true} but got {int_zero}")


def test_base_attr_verify():
    """
    Check that a BaseAttr constraint verifies an attribute with the expected
    base attribute.
    """
    eq_true_constraint = BaseAttr(BoolData)
    eq_true_constraint.verify(BoolData(True))
    eq_true_constraint.verify(BoolData(False))


def test_base_attr_verify_wrong_base_fail():
    """
    Check that a BaseAttr constraint fails to verify an attribute with a
    different base attribute.
    """
    eq_true_constraint = BaseAttr(BoolData)
    int_zero = IntData(0)
    with pytest.raises(VerifyException) as e:
        eq_true_constraint.verify(int_zero)
    assert e.value.args[0] == (
        f"{int_zero} should be of base attribute {BoolData.name}")


def test_any_attr_verify():
    """Check that an AnyAttr verifies any attribute."""
    any_constraint = AnyAttr()
    any_constraint.verify(BoolData(True))
    any_constraint.verify(BoolData(False))
    any_constraint.verify(IntData(0))


@dataclass
class LessThan(AttrConstraint):
    bound: int

    def verify(self, attr: Attribute) -> None:
        if not isinstance(attr, IntData):
            raise VerifyException(
                f"{attr} should be of base attribute {IntData.name}")
        if attr.data >= self.bound:
            raise VerifyException(
                f"{attr} should hold a value less than {self.bound}")


@dataclass
class GreaterThan(AttrConstraint):
    bound: int

    def verify(self, attr: Attribute) -> None:
        if not isinstance(attr, IntData):
            raise VerifyException(
                f"{attr} should be of base attribute {IntData.name}")
        if attr.data <= self.bound:
            raise VerifyException(
                f"{attr} should hold a value greater than {self.bound}")


def test_anyof_verify():
    """
    Check that an AnyOf constraint verifies if one of the constraints
    verify.
    """
    constraint = AnyOf([LessThan(0), GreaterThan(10)])
    constraint.verify(IntData(-1))
    constraint.verify(IntData(-10))
    constraint.verify(IntData(11))
    constraint.verify(IntData(100))


def test_anyof_verify_fail():
    """
    Check that an AnyOf constraint fails to verify if none of the constraints
    verify.
    """
    constraint = AnyOf([LessThan(0), GreaterThan(10)])

    zero = IntData(0)
    ten = IntData(10)

    with pytest.raises(VerifyException) as e:
        constraint.verify(zero)
    assert e.value.args[0] == f"Unexpected attribute {zero}"

    with pytest.raises(VerifyException) as e:
        constraint.verify(ten)
    assert e.value.args[0] == f"Unexpected attribute {ten}"


def test_allof_verify():
    """
    Check that an AllOf constraint verifies if all of the constraints
    verify.
    """
    constraint = AllOf([LessThan(10), GreaterThan(0)])
    constraint.verify(IntData(1))
    constraint.verify(IntData(9))
    constraint.verify(IntData(5))


def test_allof_verify_fail():
    """
    Check that an AllOf constraint fails to verify if one of the constraints
    fails to verify.
    """
    constraint = AllOf([LessThan(10), GreaterThan(0)])

    with pytest.raises(VerifyException) as e:
        constraint.verify(IntData(10))
    assert e.value.args[0] == f"{IntData(10)} should hold a value less than 10"

    with pytest.raises(VerifyException) as e:
        constraint.verify(IntData(0))
    assert e.value.args[
        0] == f"{IntData(0)} should hold a value greater than 0"


def test_param_attr_verify():
    bool_true = BoolData(True)
    constraint = ParamAttrConstraint(
        DoubleParamAttr, [EqAttrConstraint(bool_true),
                          BaseAttr(IntData)])
    constraint.verify(DoubleParamAttr([bool_true, IntData(0)]))
    constraint.verify(DoubleParamAttr([bool_true, IntData(42)]))


def test_param_attr_verify_base_fail():
    bool_true = BoolData(True)
    constraint = ParamAttrConstraint(
        DoubleParamAttr, [EqAttrConstraint(bool_true),
                          BaseAttr(IntData)])
    with pytest.raises(VerifyException) as e:
        constraint.verify(bool_true)
    assert e.value.args[0] == (
        f"{bool_true} should be of base attribute {DoubleParamAttr.name}")


def test_param_attr_verify_params_num_params_fail():
    bool_true = BoolData(True)
    constraint = ParamAttrConstraint(DoubleParamAttr,
                                     [EqAttrConstraint(bool_true)])
    attr = DoubleParamAttr([bool_true, IntData(0)])
    with pytest.raises(VerifyException) as e:
        constraint.verify(attr)
    assert e.value.args[0] == (f"1 parameters expected, but got 2")


def test_param_attr_verify_params_fail():
    bool_true = BoolData(True)
    bool_false = BoolData(False)
    constraint = ParamAttrConstraint(
        DoubleParamAttr, [EqAttrConstraint(bool_true),
                          BaseAttr(IntData)])

    with pytest.raises(VerifyException) as e:
        constraint.verify(DoubleParamAttr([bool_true, bool_false]))
    assert e.value.args[0] == (
        f"{bool_false} should be of base attribute {IntData.name}")

    with pytest.raises(VerifyException) as e:
        constraint.verify(DoubleParamAttr([bool_false, IntData(0)]))
    assert e.value.args[0] == (
        f"Expected attribute {bool_true} but got {bool_false}")